#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import copy

from keystoneauth1.exceptions import http as ksa_exceptions
from osc_lib import exceptions

from openstackclient.identity.v3 import registered_limit
from openstackclient.tests.unit import fakes
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestRegisteredLimit(identity_fakes.TestIdentityv3):
    def setUp(self):
        super().setUp()

        self.registered_limit_mock = self.identity_client.registered_limits

        self.services_mock = self.identity_client.services
        self.services_mock.reset_mock()

        self.regions_mock = self.identity_client.regions
        self.regions_mock.reset_mock()


class TestRegisteredLimitCreate(TestRegisteredLimit):
    def setUp(self):
        super().setUp()

        self.service = fakes.FakeResource(
            None, copy.deepcopy(identity_fakes.SERVICE), loaded=True
        )
        self.services_mock.get.return_value = self.service

        self.region = fakes.FakeResource(
            None, copy.deepcopy(identity_fakes.REGION), loaded=True
        )
        self.regions_mock.get.return_value = self.region

        self.cmd = registered_limit.CreateRegisteredLimit(self.app, None)

    def test_registered_limit_create_without_options(self):
        self.registered_limit_mock.create.return_value = fakes.FakeResource(
            None, copy.deepcopy(identity_fakes.REGISTERED_LIMIT), loaded=True
        )

        resource_name = identity_fakes.registered_limit_resource_name
        default_limit = identity_fakes.registered_limit_default_limit
        arglist = [
            '--service',
            identity_fakes.service_id,
            '--default-limit',
            '10',
            resource_name,
        ]

        verifylist = [
            ('service', identity_fakes.service_id),
            ('default_limit', default_limit),
            ('resource_name', resource_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'description': None, 'region': None}
        self.registered_limit_mock.create.assert_called_with(
            self.service, resource_name, default_limit, **kwargs
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )

        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.registered_limit_default_limit,
            None,
            identity_fakes.registered_limit_id,
            None,
            identity_fakes.registered_limit_resource_name,
            identity_fakes.service_id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_create_with_options(self):
        self.registered_limit_mock.create.return_value = fakes.FakeResource(
            None,
            copy.deepcopy(identity_fakes.REGISTERED_LIMIT_OPTIONS),
            loaded=True,
        )

        resource_name = identity_fakes.registered_limit_resource_name
        default_limit = identity_fakes.registered_limit_default_limit
        description = identity_fakes.registered_limit_description
        arglist = [
            '--region',
            identity_fakes.region_id,
            '--description',
            description,
            '--service',
            identity_fakes.service_id,
            '--default-limit',
            '10',
            resource_name,
        ]

        verifylist = [
            ('region', identity_fakes.region_id),
            ('description', description),
            ('service', identity_fakes.service_id),
            ('default_limit', default_limit),
            ('resource_name', resource_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'description': description, 'region': self.region}
        self.registered_limit_mock.create.assert_called_with(
            self.service, resource_name, default_limit, **kwargs
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )

        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.registered_limit_default_limit,
            description,
            identity_fakes.registered_limit_id,
            identity_fakes.region_id,
            identity_fakes.registered_limit_resource_name,
            identity_fakes.service_id,
        )
        self.assertEqual(datalist, data)


class TestRegisteredLimitDelete(TestRegisteredLimit):
    def setUp(self):
        super().setUp()

        self.cmd = registered_limit.DeleteRegisteredLimit(self.app, None)

    def test_registered_limit_delete(self):
        self.registered_limit_mock.delete.return_value = None

        arglist = [identity_fakes.registered_limit_id]
        verifylist = [
            ('registered_limits', [identity_fakes.registered_limit_id])
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.registered_limit_mock.delete.assert_called_with(
            identity_fakes.registered_limit_id
        )
        self.assertIsNone(result)

    def test_registered_limit_delete_with_exception(self):
        return_value = ksa_exceptions.NotFound()
        self.registered_limit_mock.delete.side_effect = return_value

        arglist = ['fake-registered-limit-id']
        verifylist = [('registered_limits', ['fake-registered-limit-id'])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual(
                '1 of 1 registered limits failed to delete.', str(e)
            )


class TestRegisteredLimitShow(TestRegisteredLimit):
    def setUp(self):
        super().setUp()

        self.registered_limit_mock.get.return_value = fakes.FakeResource(
            None, copy.deepcopy(identity_fakes.REGISTERED_LIMIT), loaded=True
        )

        self.cmd = registered_limit.ShowRegisteredLimit(self.app, None)

    def test_registered_limit_show(self):
        arglist = [identity_fakes.registered_limit_id]
        verifylist = [
            ('registered_limit_id', identity_fakes.registered_limit_id)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.registered_limit_mock.get.assert_called_with(
            identity_fakes.registered_limit_id
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.registered_limit_default_limit,
            None,
            identity_fakes.registered_limit_id,
            None,
            identity_fakes.registered_limit_resource_name,
            identity_fakes.service_id,
        )
        self.assertEqual(datalist, data)


class TestRegisteredLimitSet(TestRegisteredLimit):
    def setUp(self):
        super().setUp()
        self.cmd = registered_limit.SetRegisteredLimit(self.app, None)

    def test_registered_limit_set_description(self):
        registered_limit = copy.deepcopy(identity_fakes.REGISTERED_LIMIT)
        registered_limit['description'] = (
            identity_fakes.registered_limit_description
        )
        self.registered_limit_mock.update.return_value = fakes.FakeResource(
            None, registered_limit, loaded=True
        )

        arglist = [
            '--description',
            identity_fakes.registered_limit_description,
            identity_fakes.registered_limit_id,
        ]
        verifylist = [
            ('description', identity_fakes.registered_limit_description),
            ('registered_limit_id', identity_fakes.registered_limit_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.registered_limit_mock.update.assert_called_with(
            identity_fakes.registered_limit_id,
            service=None,
            resource_name=None,
            default_limit=None,
            description=identity_fakes.registered_limit_description,
            region=None,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.registered_limit_default_limit,
            identity_fakes.registered_limit_description,
            identity_fakes.registered_limit_id,
            None,
            identity_fakes.registered_limit_resource_name,
            identity_fakes.service_id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_set_default_limit(self):
        registered_limit = copy.deepcopy(identity_fakes.REGISTERED_LIMIT)
        default_limit = 20
        registered_limit['default_limit'] = default_limit
        self.registered_limit_mock.update.return_value = fakes.FakeResource(
            None, registered_limit, loaded=True
        )

        arglist = [
            '--default-limit',
            str(default_limit),
            identity_fakes.registered_limit_id,
        ]
        verifylist = [
            ('default_limit', default_limit),
            ('registered_limit_id', identity_fakes.registered_limit_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.registered_limit_mock.update.assert_called_with(
            identity_fakes.registered_limit_id,
            service=None,
            resource_name=None,
            default_limit=default_limit,
            description=None,
            region=None,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            default_limit,
            None,
            identity_fakes.registered_limit_id,
            None,
            identity_fakes.registered_limit_resource_name,
            identity_fakes.service_id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_set_resource_name(self):
        registered_limit = copy.deepcopy(identity_fakes.REGISTERED_LIMIT)
        resource_name = 'volumes'
        registered_limit['resource_name'] = resource_name
        self.registered_limit_mock.update.return_value = fakes.FakeResource(
            None, registered_limit, loaded=True
        )

        arglist = [
            '--resource-name',
            resource_name,
            identity_fakes.registered_limit_id,
        ]
        verifylist = [
            ('resource_name', resource_name),
            ('registered_limit_id', identity_fakes.registered_limit_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.registered_limit_mock.update.assert_called_with(
            identity_fakes.registered_limit_id,
            service=None,
            resource_name=resource_name,
            default_limit=None,
            description=None,
            region=None,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.registered_limit_default_limit,
            None,
            identity_fakes.registered_limit_id,
            None,
            resource_name,
            identity_fakes.service_id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_set_service(self):
        registered_limit = copy.deepcopy(identity_fakes.REGISTERED_LIMIT)
        service = identity_fakes.FakeService.create_one_service()
        registered_limit['service_id'] = service.id
        self.registered_limit_mock.update.return_value = fakes.FakeResource(
            None, registered_limit, loaded=True
        )
        self.services_mock.get.return_value = service

        arglist = ['--service', service.id, identity_fakes.registered_limit_id]
        verifylist = [
            ('service', service.id),
            ('registered_limit_id', identity_fakes.registered_limit_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.registered_limit_mock.update.assert_called_with(
            identity_fakes.registered_limit_id,
            service=service,
            resource_name=None,
            default_limit=None,
            description=None,
            region=None,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.registered_limit_default_limit,
            None,
            identity_fakes.registered_limit_id,
            None,
            identity_fakes.registered_limit_resource_name,
            service.id,
        )
        self.assertEqual(datalist, data)

    def test_registered_limit_set_region(self):
        registered_limit = copy.deepcopy(identity_fakes.REGISTERED_LIMIT)
        region = identity_fakes.REGION
        region['id'] = 'RegionTwo'
        region = fakes.FakeResource(None, copy.deepcopy(region), loaded=True)
        registered_limit['region_id'] = region.id
        self.registered_limit_mock.update.return_value = fakes.FakeResource(
            None, registered_limit, loaded=True
        )
        self.regions_mock.get.return_value = region

        arglist = ['--region', region.id, identity_fakes.registered_limit_id]
        verifylist = [
            ('region', region.id),
            ('registered_limit_id', identity_fakes.registered_limit_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.registered_limit_mock.update.assert_called_with(
            identity_fakes.registered_limit_id,
            service=None,
            resource_name=None,
            default_limit=None,
            description=None,
            region=region,
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.registered_limit_default_limit,
            None,
            identity_fakes.registered_limit_id,
            region.id,
            identity_fakes.registered_limit_resource_name,
            identity_fakes.service_id,
        )
        self.assertEqual(datalist, data)


class TestRegisteredLimitList(TestRegisteredLimit):
    def setUp(self):
        super().setUp()

        self.registered_limit_mock.get.return_value = fakes.FakeResource(
            None, copy.deepcopy(identity_fakes.REGISTERED_LIMIT), loaded=True
        )

        self.cmd = registered_limit.ShowRegisteredLimit(self.app, None)

    def test_limit_show(self):
        arglist = [identity_fakes.registered_limit_id]
        verifylist = [
            ('registered_limit_id', identity_fakes.registered_limit_id)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.registered_limit_mock.get.assert_called_with(
            identity_fakes.registered_limit_id
        )

        collist = (
            'default_limit',
            'description',
            'id',
            'region_id',
            'resource_name',
            'service_id',
        )
        self.assertEqual(collist, columns)
        datalist = (
            identity_fakes.registered_limit_default_limit,
            None,
            identity_fakes.registered_limit_id,
            None,
            identity_fakes.registered_limit_resource_name,
            identity_fakes.service_id,
        )
        self.assertEqual(datalist, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from osc_lib import exceptions

from openstackclient.api import compute_v2
from openstackclient.network.v2 import floating_ip as fip
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit import utils as tests_utils


@mock.patch.object(compute_v2, 'create_floating_ip')
class TestCreateFloatingIPCompute(compute_fakes.TestComputev2):
    _floating_ip = compute_fakes.create_one_floating_ip()

    columns = (
        'fixed_ip',
        'id',
        'instance_id',
        'ip',
        'pool',
    )

    data = (
        _floating_ip['fixed_ip'],
        _floating_ip['id'],
        _floating_ip['instance_id'],
        _floating_ip['ip'],
        _floating_ip['pool'],
    )

    def setUp(self):
        super().setUp()

        self.app.client_manager.network_endpoint_enabled = False

        self.cmd = fip.CreateFloatingIP(self.app, None)

    def test_floating_ip_create_no_arg(self, fip_mock):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_floating_ip_create_default(self, fip_mock):
        fip_mock.return_value = self._floating_ip
        arglist = [
            self._floating_ip['pool'],
        ]
        verifylist = [
            ('network', self._floating_ip['pool']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        fip_mock.assert_called_once_with(
            self.compute_client, self._floating_ip['pool']
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


@mock.patch.object(compute_v2, 'delete_floating_ip')
class TestDeleteFloatingIPCompute(compute_fakes.TestComputev2):
    _floating_ips = compute_fakes.create_floating_ips(count=2)

    def setUp(self):
        super().setUp()

        self.app.client_manager.network_endpoint_enabled = False

        self.cmd = fip.DeleteFloatingIP(self.app, None)

    def test_floating_ip_delete(self, fip_mock):
        fip_mock.return_value = mock.Mock(return_value=None)
        arglist = [
            self._floating_ips[0]['id'],
        ]
        verifylist = [
            ('floating_ip', [self._floating_ips[0]['id']]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        fip_mock.assert_called_once_with(
            self.compute_client, self._floating_ips[0]['id']
        )
        self.assertIsNone(result)

    def test_floating_ip_delete_multi(self, fip_mock):
        fip_mock.return_value = mock.Mock(return_value=None)
        arglist = [
            self._floating_ips[0]['id'],
            self._floating_ips[1]['id'],
        ]
        verifylist = [
            ('floating_ip', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        fip_mock.assert_has_calls(
            [
                mock.call(self.compute_client, self._floating_ips[0]['id']),
                mock.call(self.compute_client, self._floating_ips[1]['id']),
            ]
        )
        self.assertIsNone(result)

    def test_floating_ip_delete_multi_exception(self, fip_mock):
        fip_mock.return_value = mock.Mock(return_value=None)
        fip_mock.side_effect = [
            mock.Mock(return_value=None),
            exceptions.CommandError,
        ]
        arglist = [
            self._floating_ips[0]['id'],
            'unexist_floating_ip',
        ]
        verifylist = [
            (
                'floating_ip',
                [self._floating_ips[0]['id'], 'unexist_floating_ip'],
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 floating_ips failed to delete.', str(e))

        fip_mock.assert_any_call(
            self.compute_client, self._floating_ips[0]['id']
        )
        fip_mock.assert_any_call(self.compute_client, 'unexist_floating_ip')


@mock.patch.object(compute_v2, 'list_floating_ips')
class TestListFloatingIPCompute(compute_fakes.TestComputev2):
    _floating_ips = compute_fakes.create_floating_ips(count=3)

    columns = (
        'ID',
        'Floating IP Address',
        'Fixed IP Address',
        'Server',
        'Pool',
    )

    data = []
    for ip in _floating_ips:
        data.append(
            (
                ip['id'],
                ip['ip'],
                ip['fixed_ip'],
                ip['instance_id'],
                ip['pool'],
            )
        )

    def setUp(self):
        super().setUp()

        self.app.client_manager.network_endpoint_enabled = False

        self.cmd = fip.ListFloatingIP(self.app, None)

    def test_floating_ip_list(self, fip_mock):
        fip_mock.return_value = self._floating_ips
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        fip_mock.assert_called_once_with(self.compute_client)
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


@mock.patch.object(compute_v2, 'get_floating_ip')
class TestShowFloatingIPCompute(compute_fakes.TestComputev2):
    _floating_ip = compute_fakes.create_one_floating_ip()

    columns = (
        'fixed_ip',
        'id',
        'instance_id',
        'ip',
        'pool',
    )

    data = (
        _floating_ip['fixed_ip'],
        _floating_ip['id'],
        _floating_ip['instance_id'],
        _floating_ip['ip'],
        _floating_ip['pool'],
    )

    def setUp(self):
        super().setUp()

        self.app.client_manager.network_endpoint_enabled = False

        self.cmd = fip.ShowFloatingIP(self.app, None)

    def test_floating_ip_show(self, fip_mock):
        fip_mock.return_value = self._floating_ip
        arglist = [
            self._floating_ip['id'],
        ]
        verifylist = [
            ('floating_ip', self._floating_ip['id']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        fip_mock.assert_called_once_with(
            self.compute_client, self._floating_ip['id']
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Auto-allocated Topology Implementations"""

import logging

from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common

LOG = logging.getLogger(__name__)


def _get_columns(item):
    hidden_columns = ['name', 'location', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


def _format_check_resource_columns():
    return ('dry_run',)


def _format_check_resource(item):
    item_id = getattr(item, 'id', False)
    if item_id == 'dry-run=pass':
        item.check_resource = 'pass'
    return item


def _get_attrs(client_manager, parsed_args):
    attrs = {}
    if parsed_args.project:
        identity_client = client_manager.identity
        project_id = identity_common.find_project(
            identity_client,
            parsed_args.project,
            parsed_args.project_domain,
        ).id
        attrs['project_id'] = project_id
    if parsed_args.check_resources:
        attrs['check_resources'] = True

    return attrs


# TODO(ankur-gupta-f): Use the SDK resource mapped attribute names once the
# OSC minimum requirements include SDK 1.0.
class CreateAutoAllocatedTopology(command.ShowOne):
    _description = _("Create the  auto allocated topology for project")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "Return the auto allocated topology for a given project. "
                "Default is current project."
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--check-resources',
            action='store_true',
            help=_(
                "Validate the requirements for auto allocated topology. "
                "Does not return a topology."
            ),
        )
        parser.add_argument(
            '--or-show',
            action='store_true',
            default=True,
            help=_(
                "If topology exists returns the topology's "
                "information (default)"
            ),
        )

        return parser

    def check_resource_topology(self, client, parsed_args):
        obj = client.validate_auto_allocated_topology(parsed_args.project)

        columns = _format_check_resource_columns()
        data = utils.get_item_properties(
            _format_check_resource(obj), columns, formatters={}
        )

        return (columns, data)

    def get_topology(self, client, parsed_args):
        obj = client.get_auto_allocated_topology(parsed_args.project)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns, formatters={})
        return (display_columns, data)

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        if parsed_args.check_resources:
            columns, data = self.check_resource_topology(client, parsed_args)
        else:
            columns, data = self.get_topology(client, parsed_args)
        return (columns, data)


# TODO(ankur-gupta-f): Use the SDK resource mapped attribute names once the
# OSC minimum requirements include SDK 1.0.
class DeleteAutoAllocatedTopology(command.Command):
    _description = _("Delete auto allocated topology for project")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                'Delete auto allocated topology for a given project. '
                'Default is the current project.'
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)

        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        client.delete_auto_allocated_topology(parsed_args.project)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstackclient.common import availability_zone
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes


def _build_compute_az_datalist(compute_az, long_datalist=False):
    datalist = ()
    if not long_datalist:
        datalist = (
            compute_az.name,
            'available',
        )
    else:
        for host, services in compute_az.hosts.items():
            for service, state in services.items():
                datalist += (
                    compute_az.name,
                    'available',
                    '',
                    host,
                    service,
                    'enabled :-) ' + state['updated_at'],
                )
    return (datalist,)


def _build_volume_az_datalist(volume_az, long_datalist=False):
    datalist = ()
    if not long_datalist:
        datalist = (
            volume_az.name,
            'available',
        )
    else:
        datalist = (
            volume_az.name,
            'available',
            '',
            '',
            '',
            '',
        )
    return (datalist,)


def _build_network_az_datalist(network_az, long_datalist=False):
    datalist = ()
    if not long_datalist:
        datalist = (
            network_az.name,
            network_az.state,
        )
    else:
        datalist = (
            network_az.name,
            network_az.state,
            network_az.resource,
            '',
            '',
            '',
        )
    return (datalist,)


class TestAvailabilityZoneList(
    network_fakes.FakeClientMixin,
    volume_fakes.FakeClientMixin,
    compute_fakes.FakeClientMixin,
    utils.TestCommand,
):
    compute_azs = compute_fakes.create_availability_zones()
    volume_azs = volume_fakes.create_availability_zones(count=1)
    network_azs = network_fakes.create_availability_zones()

    short_columnslist = ('Zone Name', 'Zone Status')
    long_columnslist = (
        'Zone Name',
        'Zone Status',
        'Zone Resource',
        'Host Name',
        'Service Name',
        'Service Status',
    )

    def setUp(self):
        super().setUp()

        self.compute_client.availability_zones.return_value = self.compute_azs
        self.volume_sdk_client.availability_zones.return_value = (
            self.volume_azs
        )
        self.network_client.availability_zones.return_value = self.network_azs

        # Get the command object to test
        self.cmd = availability_zone.ListAvailabilityZone(self.app, None)

    def test_availability_zone_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.availability_zones.assert_called_with(details=True)
        self.volume_sdk_client.availability_zones.assert_called_with()
        self.network_client.availability_zones.assert_called_with()

        self.assertEqual(self.short_columnslist, columns)
        datalist = ()
        for compute_az in self.compute_azs:
            datalist += _build_compute_az_datalist(compute_az)
        for volume_az in self.volume_azs:
            datalist += _build_volume_az_datalist(volume_az)
        for network_az in self.network_azs:
            datalist += _build_network_az_datalist(network_az)
        self.assertEqual(datalist, tuple(data))

    def test_availability_zone_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.availability_zones.assert_called_with(details=True)
        self.volume_sdk_client.availability_zones.assert_called_with()
        self.network_client.availability_zones.assert_called_with()

        self.assertEqual(self.long_columnslist, columns)
        datalist = ()
        for compute_az in self.compute_azs:
            datalist += _build_compute_az_datalist(
                compute_az, long_datalist=True
            )
        for volume_az in self.volume_azs:
            datalist += _build_volume_az_datalist(
                volume_az, long_datalist=True
            )
        for network_az in self.network_azs:
            datalist += _build_network_az_datalist(
                network_az, long_datalist=True
            )
        self.assertEqual(datalist, tuple(data))

    def test_availability_zone_list_compute(self):
        arglist = [
            '--compute',
        ]
        verifylist = [
            ('compute', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.availability_zones.assert_called_with(details=True)
        self.volume_sdk_client.availability_zones.assert_not_called()
        self.network_client.availability_zones.assert_not_called()

        self.assertEqual(self.short_columnslist, columns)
        datalist = ()
        for compute_az in self.compute_azs:
            datalist += _build_compute_az_datalist(compute_az)
        self.assertEqual(datalist, tuple(data))

    def test_availability_zone_list_volume(self):
        arglist = [
            '--volume',
        ]
        verifylist = [
            ('volume', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.availability_zones.assert_not_called()
        self.volume_sdk_client.availability_zones.assert_called_with()
        self.network_client.availability_zones.assert_not_called()

        self.assertEqual(self.short_columnslist, columns)
        datalist = ()
        for volume_az in self.volume_azs:
            datalist += _build_volume_az_datalist(volume_az)
        self.assertEqual(datalist, tuple(data))

    def test_availability_zone_list_network(self):
        arglist = [
            '--network',
        ]
        verifylist = [
            ('network', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.compute_client.availability_zones.assert_not_called()
        self.volume_sdk_client.availability_zones.assert_not_called()
        self.network_client.availability_zones.assert_called_with()

        self.assertEqual(self.short_columnslist, columns)
        datalist = ()
        for network_az in self.network_azs:
            datalist += _build_network_az_datalist(network_az)
        self.assertEqual(datalist, tuple(data))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from tempest.lib.common.utils import data_utils
from tempest.lib import exceptions

from openstackclient.tests.functional import base


class QuotaTests(base.TestCase):
    """Functional tests for quota

    Note that for 'set' tests use different quotas for each API in different
    test runs as these may run in parallel and otherwise step on each other.
    """

    PROJECT_NAME: str

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.haz_network = cls.is_service_enabled('network')
        cls.PROJECT_NAME = data_utils.rand_name('TestProject')
        cls.openstack(f'project create {cls.PROJECT_NAME}')

    @classmethod
    def tearDownClass(cls):
        cls.openstack(f'project delete {cls.PROJECT_NAME}')
        super().tearDownClass()

    def test_quota_list_network_option(self):
        if not self.haz_network:
            self.skipTest("No Network service present")
        self.openstack('quota set --networks 40 ' + self.PROJECT_NAME)
        cmd_output = self.openstack(
            'quota list --network',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        self.assertEqual(
            40,
            cmd_output[0]["Networks"],
        )

    def test_quota_list_compute_option(self):
        self.openstack('quota set --instances 30 ' + self.PROJECT_NAME)
        cmd_output = self.openstack(
            'quota list --compute',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        self.assertEqual(
            30,
            cmd_output[0]["Instances"],
        )

    def test_quota_list_volume_option(self):
        self.openstack('quota set --volumes 20 ' + self.PROJECT_NAME)
        cmd_output = self.openstack(
            'quota list --volume',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        self.assertEqual(
            20,
            cmd_output[0]["Volumes"],
        )

    def test_quota_set_project(self):
        """Test quota set, show"""
        network_option = ""
        if self.haz_network:
            network_option = "--routers 21 "
        self.openstack(
            'quota set --cores 31 --backups 41 '
            + network_option
            + self.PROJECT_NAME
        )
        cmd_output = self.openstack(
            'quota show ' + self.PROJECT_NAME,
            parse_output=True,
        )
        cmd_output = {x['Resource']: x['Limit'] for x in cmd_output}
        self.assertIsNotNone(cmd_output)
        self.assertEqual(
            31,
            cmd_output["cores"],
        )
        self.assertEqual(
            41,
            cmd_output["backups"],
        )
        if self.haz_network:
            self.assertEqual(
                21,
                cmd_output["routers"],
            )

        # Check default quotas
        cmd_output = self.openstack(
            'quota show --default',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        # We don't necessarily know the default quotas, we're checking the
        # returned attributes
        cmd_output = {x['Resource']: x['Limit'] for x in cmd_output}
        self.assertTrue(cmd_output["cores"] >= 0)
        self.assertTrue(cmd_output["backups"] >= 0)
        if self.haz_network:
            self.assertTrue(cmd_output["routers"] >= 0)

    def test_quota_set_default(self):
        self.openstack(
            'quota set --key-pairs 33 --snapshots 43 --class default'
        )
        cmd_output = self.openstack(
            'quota show --default',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        cmd_output = {x['Resource']: x['Limit'] for x in cmd_output}
        self.assertEqual(33, cmd_output["key-pairs"])
        self.assertEqual(43, cmd_output["snapshots"])

    def _restore_quota_limit(self, resource, limit, project):
        self.openstack(f'quota set --{resource} {limit} {project}')

    def test_quota_set_network(self):
        if not self.haz_network:
            self.skipTest('No Network service present')
        if not self.is_extension_enabled('quota-check-limit'):
            self.skipTest('No "quota-check-limit" extension present')

        cmd_output = self.openstack(
            'quota list --network',
            parse_output=True,
        )
        self.addCleanup(
            self._restore_quota_limit,
            'network',
            cmd_output[0]['Networks'],
            self.PROJECT_NAME,
        )

        self.openstack('quota set --networks 40 ' + self.PROJECT_NAME)
        cmd_output = self.openstack(
            'quota list --network',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        self.assertEqual(40, cmd_output[0]['Networks'])

        # That will ensure we have at least two networks in the system.
        for _ in range(2):
            self.openstack(
                f'network create --project {self.PROJECT_NAME} {uuid.uuid4().hex}'
            )

        self.assertRaises(
            exceptions.CommandFailed,
            self.openstack,
            'quota set --networks 1 ' + self.PROJECT_NAME,
        )

    def test_quota_set_network_with_force(self):
        self.skipTest('story 2010110')
        if not self.haz_network:
            self.skipTest('No Network service present')
        # NOTE(ralonsoh): the Neutron support for the flag "check-limit" was
        # added with the extension "quota-check-limit". The flag "force" was
        # added too in order to change the behaviour of Neutron quota engine
        # and mimic the Nova one: by default the engine will check the resource
        # usage before setting the new limit; with "force", this check will be
        # skipped (in Yoga, this behaviour is still NOT the default in
        # Neutron).
        if not self.is_extension_enabled('quota-check-limit'):
            self.skipTest('No "quota-check-limit" extension present')

        cmd_output = self.openstack(
            'quota list --network',
            parse_output=True,
        )
        self.addCleanup(
            self._restore_quota_limit,
            'network',
            cmd_output[0]['Networks'],
            self.PROJECT_NAME,
        )

        self.openstack('quota set --networks 40 ' + self.PROJECT_NAME)
        cmd_output = self.openstack(
            'quota list --network',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        self.assertEqual(40, cmd_output[0]['Networks'])

        # That will ensure we have at least two networks in the system.
        for _ in range(2):
            self.openstack(
                f'network create --project {self.PROJECT_NAME} {uuid.uuid4().hex}'
            )

        self.openstack('quota set --networks 1 --force ' + self.PROJECT_NAME)
        cmd_output = self.openstack(
            'quota list --network',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        self.assertEqual(1, cmd_output[0]['Networks'])

    def test_quota_show(self):
        expected_headers = ["Resource", "Limit"]
        cmd_output = self.openstack(
            'quota show',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        resources = []
        for row in cmd_output:
            row_headers = [str(r) for r in row.keys()]
            self.assertEqual(sorted(expected_headers), sorted(row_headers))
            resources.append(row['Resource'])
        # Ensure that returned quota has network quota...
        self.assertIn("networks", resources)
        # ...and compute quota
        self.assertIn("instances", resources)

    def test_quota_show_usage_option(self):
        expected_headers = ["Resource", "Limit", "In Use", "Reserved"]
        cmd_output = self.openstack(
            'quota show --usage',
            parse_output=True,
        )
        self.assertIsNotNone(cmd_output)
        resources = []
        for row in cmd_output:
            row_headers = [str(r) for r in row.keys()]
            self.assertEqual(sorted(expected_headers), sorted(row_headers))
            resources.append(row['Resource'])
            for header in expected_headers[1:]:
                self.assertIsInstance(row[header], int)
        # Ensure that returned quota has network quota...
        self.assertIn("networks", resources)
        # ...and compute quota
        self.assertIn("instances", resources)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Network Service Providers Implementation"""

from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


class ListNetworkServiceProvider(command.Lister):
    _description = _("List Service Providers")

    def take_action(self, parsed_args):
        client = self.app.client_manager.network

        columns = (
            "service_type",
            "name",
            "is_default",
        )
        column_headers = (
            "Service Type",
            "Name",
            "Default",
        )

        data = client.service_providers()
        return (
            column_headers,
            (
                utils.get_item_properties(
                    s,
                    columns,
                )
                for s in data
            ),
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from openstackclient.tests.functional.identity.v3 import common


class CatalogTests(common.IdentityTests):
    """Functional tests for catalog commands"""

    def test_catalog(self):
        """Test catalog list and show functionality"""
        # Create a test service for isolated testing
        _dummy_service_name = self._create_dummy_service(add_clean_up=True)

        # list catalogs
        raw_output = self.openstack('catalog list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, ['Name', 'Type', 'Endpoints'])

        # Verify created service appears in catalog
        service_names = [
            item.get('Name') for item in items if item.get('Name')
        ]
        self.assertIn(
            _dummy_service_name,
            service_names,
            "Created dummy service should be present in catalog",
        )

        # show service (by name)
        raw_output = self.openstack(f'catalog show {_dummy_service_name}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, ['endpoints', 'name', 'type', 'id'])

        # Extract the type from the dummy service
        _dummy_service_type = next(
            (item['type'] for item in items if 'type' in item), None
        )

        # show service (by type)
        raw_output = self.openstack(f'catalog show {_dummy_service_type}')
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, ['endpoints', 'name', 'type', 'id'])

        # show service (non-existent)
        result = self.openstack(
            'catalog show nonexistent-service-xyz', fail_ok=True
        )
        self.assertEqual(
            '',
            result.strip(),
            "Non-existent service should return empty result",
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstack.identity.v3 import domain as _domain
from openstack.test import fakes as sdk_fakes
from openstackclient.identity.v3 import domain
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes


class TestDomainCreate(identity_fakes.TestIdentityv3):
    columns = (
        'id',
        'name',
        'enabled',
        'description',
        'options',
    )

    def setUp(self):
        super().setUp()

        self.domain = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.identity_sdk_client.create_domain.return_value = self.domain
        self.datalist = (
            self.domain.id,
            self.domain.name,
            self.domain.is_enabled,
            self.domain.description,
            self.domain.options,
        )

        # Get the command object to test
        self.cmd = domain.CreateDomain(self.app, None)

    def test_domain_create_no_options(self):
        arglist = [
            self.domain.name,
        ]
        verifylist = [
            ('name', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.domain.name,
            'description': None,
            'options': {},
            'is_enabled': True,
        }
        self.identity_sdk_client.create_domain.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_domain_create_description(self):
        arglist = [
            '--description',
            'new desc',
            self.domain.name,
        ]
        verifylist = [
            ('description', 'new desc'),
            ('name', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.domain.name,
            'description': 'new desc',
            'options': {},
            'is_enabled': True,
        }
        self.identity_sdk_client.create_domain.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_domain_create_enable(self):
        arglist = [
            '--enable',
            self.domain.name,
        ]
        verifylist = [
            ('is_enabled', True),
            ('name', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.domain.name,
            'description': None,
            'options': {},
            'is_enabled': True,
        }
        self.identity_sdk_client.create_domain.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_domain_create_disable(self):
        arglist = [
            '--disable',
            self.domain.name,
        ]
        verifylist = [
            ('is_enabled', False),
            ('name', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.domain.name,
            'description': None,
            'options': {},
            'is_enabled': False,
        }
        self.identity_sdk_client.create_domain.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_domain_create_with_immutable(self):
        arglist = [
            '--immutable',
            self.domain.name,
        ]
        verifylist = [
            ('immutable', True),
            ('name', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.domain.name,
            'description': None,
            'options': {'immutable': True},
            'is_enabled': True,
        }
        self.identity_sdk_client.create_domain.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)

    def test_domain_create_with_no_immutable(self):
        arglist = [
            '--no-immutable',
            self.domain.name,
        ]
        verifylist = [
            ('immutable', False),
            ('name', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': self.domain.name,
            'description': None,
            'options': {'immutable': False},
            'is_enabled': True,
        }
        self.identity_sdk_client.create_domain.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)


class TestDomainDelete(identity_fakes.TestIdentityv3):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    def setUp(self):
        super().setUp()

        # This is the return value for utils.find_resource()
        self.identity_sdk_client.find_domain.return_value = self.domain
        self.identity_sdk_client.delete_domain.return_value = None

        # Get the command object to test
        self.cmd = domain.DeleteDomain(self.app, None)

    def test_domain_delete(self):
        arglist = [
            self.domain.id,
        ]
        verifylist = [
            ('domain', [self.domain.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_domain.assert_called_with(
            self.domain.id,
        )
        self.assertIsNone(result)


class TestDomainList(identity_fakes.TestIdentityv3):
    domain = sdk_fakes.generate_fake_resource(
        resource_type=_domain.Domain, is_enabled=True
    )
    columns = (
        'ID',
        'Name',
        'Enabled',
        'Description',
    )

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.domains.return_value = [self.domain]
        self.datalist = (
            (
                self.domain.id,
                self.domain.name,
                self.domain.is_enabled,
                self.domain.description,
            ),
        )

        # Get the command object to test
        self.cmd = domain.ListDomain(self.app, None)

    def test_domain_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.domains.assert_called_with()

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_domain_list_with_option_name(self):
        arglist = ['--name', self.domain.name]
        verifylist = [('name', self.domain.name)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'name': self.domain.name}
        self.identity_sdk_client.domains.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))

    def test_domain_list_with_option_enabled(self):
        arglist = ['--enabled']
        verifylist = [('is_enabled', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'is_enabled': True}
        self.identity_sdk_client.domains.assert_called_with(**kwargs)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, tuple(data))


class TestDomainSet(identity_fakes.TestIdentityv3):
    domain = sdk_fakes.generate_fake_resource(_domain.Domain)

    def setUp(self):
        super().setUp()

        self.identity_sdk_client.find_domain.return_value = self.domain

        self.identity_sdk_client.update_domain.return_value = self.domain

        # Get the command object to test
        self.cmd = domain.SetDomain(self.app, None)

    def test_domain_set_no_options(self):
        arglist = [
            self.domain.name,
        ]
        verifylist = [
            ('domain', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        kwargs = {}
        self.identity_sdk_client.update_domain.assert_called_with(
            self.domain.id, **kwargs
        )
        self.assertIsNone(result)

    def test_domain_set_name(self):
        arglist = [
            '--name',
            'qwerty',
            self.domain.id,
        ]
        verifylist = [
            ('name', 'qwerty'),
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'name': 'qwerty',
        }
        self.identity_sdk_client.update_domain.assert_called_with(
            self.domain.id, **kwargs
        )
        self.assertIsNone(result)

    def test_domain_set_description(self):
        arglist = [
            '--description',
            'new desc',
            self.domain.id,
        ]
        verifylist = [
            ('description', 'new desc'),
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': 'new desc',
        }
        self.identity_sdk_client.update_domain.assert_called_with(
            self.domain.id, **kwargs
        )
        self.assertIsNone(result)

    def test_domain_set_enable(self):
        arglist = [
            '--enable',
            self.domain.id,
        ]
        verifylist = [
            ('is_enabled', True),
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
        }
        self.identity_sdk_client.update_domain.assert_called_with(
            self.domain.id, **kwargs
        )
        self.assertIsNone(result)

    def test_domain_set_disable(self):
        arglist = [
            '--disable',
            self.domain.id,
        ]
        verifylist = [
            ('is_enabled', False),
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': False,
        }
        self.identity_sdk_client.update_domain.assert_called_with(
            self.domain.id, **kwargs
        )
        self.assertIsNone(result)

    def test_domain_set_immutable_option(self):
        arglist = [
            '--immutable',
            self.domain.id,
        ]
        verifylist = [
            ('immutable', True),
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'options': {'immutable': True},
        }
        self.identity_sdk_client.update_domain.assert_called_with(
            self.domain.id, **kwargs
        )
        self.assertIsNone(result)

    def test_domain_set_no_immutable_option(self):
        arglist = [
            '--no-immutable',
            self.domain.id,
        ]
        verifylist = [
            ('immutable', False),
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'options': {'immutable': False},
        }
        self.identity_sdk_client.update_domain.assert_called_with(
            self.domain.id, **kwargs
        )
        self.assertIsNone(result)


class TestDomainShow(identity_fakes.TestIdentityv3):
    columns = (
        'id',
        'name',
        'enabled',
        'description',
        'options',
    )

    def setUp(self):
        super().setUp()

        self.domain = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.identity_sdk_client.find_domain.return_value = self.domain
        self.datalist = (
            self.domain.id,
            self.domain.name,
            self.domain.is_enabled,
            self.domain.description,
            self.domain.options,
        )

        # Get the command object to test
        self.cmd = domain.ShowDomain(self.app, None)

    def test_domain_show(self):
        arglist = [
            self.domain.id,
        ]
        verifylist = [
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.identity_client.tokens.get_token_data.return_value = {
            'token': {'project': {'domain': {'id': 'd1', 'name': 'd1'}}}
        }

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.find_domain.assert_called_with(
            self.domain.id, ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.datalist, data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import fixtures

from openstackclient.tests.functional.image import base


class ImageTests(base.BaseImageTests):
    """Functional tests for Image commands"""

    def setUp(self):
        super().setUp()

        if not self.haz_v1_api:
            self.skipTest('No Image v1 API present')

        ver_fixture = fixtures.EnvironmentVariable('OS_IMAGE_API_VERSION', '1')
        self.useFixture(ver_fixture)

        self.name = uuid.uuid4().hex
        output = self.openstack(
            'image create ' + self.name,
            parse_output=True,
        )
        self.image_id = output["id"]
        self.assertOutput(self.name, output['name'])

    def tearDown(self):
        try:
            self.openstack('image delete ' + self.image_id)
        finally:
            super().tearDown()

    def test_image_list(self):
        output = self.openstack('image list')
        self.assertIn(self.name, [img['Name'] for img in output])

    def test_image_attributes(self):
        """Test set, unset, show on attributes, tags and properties"""

        # Test explicit attributes
        self.openstack(
            'image set '
            + '--min-disk 4 '
            + '--min-ram 5 '
            + '--disk-format qcow2 '
            + '--public '
            + self.name
        )
        output = self.openstack(
            'image show ' + self.name,
            parse_output=True,
        )
        self.assertEqual(
            4,
            output["min_disk"],
        )
        self.assertEqual(
            5,
            output["min_ram"],
        )
        self.assertEqual(
            'qcow2',
            output['disk_format'],
        )
        self.assertTrue(
            output["is_public"],
        )

        # Test properties
        self.openstack(
            'image set '
            + '--property a=b '
            + '--property c=d '
            + '--public '
            + self.name
        )
        output = self.openstack(
            'image show ' + self.name,
            parse_output=True,
        )
        self.assertEqual(
            {'a': 'b', 'c': 'd'},
            output["properties"],
        )



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.network import common

RULE_TYPE_BANDWIDTH_LIMIT = 'bandwidth-limit'
RULE_TYPE_DSCP_MARKING = 'dscp-marking'
RULE_TYPE_MINIMUM_BANDWIDTH = 'minimum-bandwidth'
RULE_TYPE_MINIMUM_PACKET_RATE = 'minimum-packet-rate'
MANDATORY_PARAMETERS = {
    RULE_TYPE_MINIMUM_BANDWIDTH: {'min_kbps', 'direction'},
    RULE_TYPE_MINIMUM_PACKET_RATE: {'min_kpps', 'direction'},
    RULE_TYPE_DSCP_MARKING: {'dscp_mark'},
    RULE_TYPE_BANDWIDTH_LIMIT: {'max_kbps'},
}
OPTIONAL_PARAMETERS = {
    RULE_TYPE_MINIMUM_BANDWIDTH: set(),
    RULE_TYPE_MINIMUM_PACKET_RATE: set(),
    RULE_TYPE_DSCP_MARKING: set(),
    RULE_TYPE_BANDWIDTH_LIMIT: {'direction', 'max_burst_kbps'},
}
DIRECTION_EGRESS = 'egress'
DIRECTION_INGRESS = 'ingress'
DIRECTION_ANY = 'any'
DSCP_VALID_MARKS = [
    0,
    8,
    10,
    12,
    14,
    16,
    18,
    20,
    22,
    24,
    26,
    28,
    30,
    32,
    34,
    36,
    38,
    40,
    46,
    48,
    56,
]

ACTION_CREATE = 'create'
ACTION_DELETE = 'delete'
ACTION_FIND = 'find'
ACTION_SET = 'update'
ACTION_SHOW = 'get'


def _get_columns(item):
    hidden_columns = ['location', 'name', 'tenant_id']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


def _check_type_parameters(attrs, type, is_create):
    req_params = MANDATORY_PARAMETERS[type]
    opt_params = OPTIONAL_PARAMETERS[type]
    type_params = req_params | opt_params
    notreq_params = set(
        itertools.chain(
            *[v for k, v in MANDATORY_PARAMETERS.items() if k != type]
        )
    )
    notreq_params -= type_params
    if is_create and None in map(attrs.get, req_params):
        msg = _(
            '"Create" rule command for type "%(rule_type)s" requires '
            'arguments: %(args)s'
        ) % {'rule_type': type, 'args': ", ".join(sorted(req_params))}
        raise exceptions.CommandError(msg)
    if set(attrs.keys()) & notreq_params:
        msg = _(
            'Rule type "%(rule_type)s" only requires arguments: %(args)s'
        ) % {'rule_type': type, 'args': ", ".join(sorted(type_params))}
        raise exceptions.CommandError(msg)


def _get_attrs(network_client, parsed_args, is_create=False):
    attrs = {}
    qos = network_client.find_qos_policy(
        parsed_args.qos_policy, ignore_missing=False
    )
    attrs['qos_policy_id'] = qos.id
    if not is_create:
        attrs['id'] = parsed_args.id
        rule_type = _find_rule_type(qos, parsed_args.id)
        if not rule_type:
            msg = _('Rule ID %(rule_id)s not found') % {
                'rule_id': parsed_args.id
            }
            raise exceptions.CommandError(msg)
    else:
        rule_type = parsed_args.type
    if parsed_args.max_kbps is not None:
        attrs['max_kbps'] = parsed_args.max_kbps
    if parsed_args.max_burst_kbits is not None:
        # NOTE(ralonsoh): this parameter must be changed in SDK and then in
        #                 Neutron API, from 'max_burst_kbps' to
        #                 'max_burst_kbits'
        attrs['max_burst_kbps'] = parsed_args.max_burst_kbits
    if parsed_args.dscp_mark is not None:
        attrs['dscp_mark'] = parsed_args.dscp_mark
    if parsed_args.min_kbps is not None:
        attrs['min_kbps'] = parsed_args.min_kbps
    if parsed_args.min_kpps is not None:
        attrs['min_kpps'] = parsed_args.min_kpps
    if parsed_args.ingress:
        attrs['direction'] = DIRECTION_INGRESS
    if parsed_args.egress:
        attrs['direction'] = DIRECTION_EGRESS
    if parsed_args.any:
        if rule_type == RULE_TYPE_MINIMUM_PACKET_RATE:
            attrs['direction'] = DIRECTION_ANY
        else:
            msg = _(
                'Direction "any" can only be used with '
                '%(rule_type_min_pps)s rule type'
            ) % {'rule_type_min_pps': RULE_TYPE_MINIMUM_PACKET_RATE}
            raise exceptions.CommandError(msg)
    _check_type_parameters(attrs, rule_type, is_create)
    return attrs


def _rule_action_call(client, action, rule_type):
    rule_type = rule_type.replace('-', '_')
    func_name = f'{action}_qos_{rule_type}_rule'
    return getattr(client, func_name)


def _find_rule_type(qos, rule_id):
    for rule in (r for r in qos.rules if r['id'] == rule_id):
        return rule['type'].replace('_', '-')
    return None


def _add_rule_arguments(parser):
    parser.add_argument(
        '--max-kbps',
        dest='max_kbps',
        metavar='<max-kbps>',
        type=int,
        help=_('Maximum bandwidth in kbps'),
    )
    parser.add_argument(
        '--max-burst-kbits',
        dest='max_burst_kbits',
        metavar='<max-burst-kbits>',
        type=int,
        help=_(
            'Maximum burst in kilobits, 0 or not specified means '
            'automatic, which is 80%% of the bandwidth limit, which works '
            'for typical TCP traffic. For details check the QoS user '
            'workflow.'
        ),
    )
    parser.add_argument(
        '--dscp-mark',
        dest='dscp_mark',
        metavar='<dscp-mark>',
        type=int,
        help=_(
            'DSCP mark: value can be 0, even numbers from 8-56, '
            'excluding 42, 44, 50, 52, and 54'
        ),
    )
    parser.add_argument(
        '--min-kbps',
        dest='min_kbps',
        metavar='<min-kbps>',
        type=int,
        help=_('Minimum guaranteed bandwidth in kbps'),
    )
    parser.add_argument(
        '--min-kpps',
        dest='min_kpps',
        metavar='<min-kpps>',
        type=int,
        help=_('Minimum guaranteed packet rate in kpps'),
    )
    direction_group = parser.add_mutually_exclusive_group()
    direction_group.add_argument(
        '--ingress',
        action='store_true',
        help=_("Ingress traffic direction from the project point of view"),
    )
    direction_group.add_argument(
        '--egress',
        action='store_true',
        help=_("Egress traffic direction from the project point of view"),
    )
    direction_group.add_argument(
        '--any',
        action='store_true',
        help=_(
            "Any traffic direction from the project point of view. Can be "
            "used only with minimum packet rate rule."
        ),
    )


class CreateNetworkQosRule(
    command.ShowOne, common.NeutronCommandWithExtraArgs
):
    _description = _("Create new Network QoS rule")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_policy',
            metavar='<qos-policy>',
            help=_('QoS policy that contains the rule (name or ID)'),
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            required=True,
            choices=[
                RULE_TYPE_MINIMUM_BANDWIDTH,
                RULE_TYPE_MINIMUM_PACKET_RATE,
                RULE_TYPE_DSCP_MARKING,
                RULE_TYPE_BANDWIDTH_LIMIT,
            ],
            help=(
                _('QoS rule type (%s)')
                % ", ".join(MANDATORY_PARAMETERS.keys())
            ),
        )
        _add_rule_arguments(parser)
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        try:
            attrs = _get_attrs(network_client, parsed_args, is_create=True)
            attrs.update(
                self._parse_extra_properties(parsed_args.extra_properties)
            )
            obj = _rule_action_call(
                network_client, ACTION_CREATE, parsed_args.type
            )(attrs.pop('qos_policy_id'), **attrs)
        except Exception as e:
            msg = _('Failed to create Network QoS rule: %(e)s') % {'e': e}
            raise exceptions.CommandError(msg)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return display_columns, data


class DeleteNetworkQosRule(command.Command):
    _description = _("Delete Network QoS rule")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_policy',
            metavar='<qos-policy>',
            help=_('QoS policy that contains the rule (name or ID)'),
        )
        parser.add_argument(
            'id',
            metavar='<rule-id>',
            help=_('Network QoS rule to delete (ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        rule_id = parsed_args.id
        try:
            qos = network_client.find_qos_policy(
                parsed_args.qos_policy, ignore_missing=False
            )
            rule_type = _find_rule_type(qos, rule_id)
            if not rule_type:
                raise Exception(f'Rule {rule_id} not found')
            _rule_action_call(network_client, ACTION_DELETE, rule_type)(
                rule_id, qos.id
            )
        except Exception as e:
            msg = _(
                'Failed to delete Network QoS rule ID "%(rule)s": %(e)s'
            ) % {'rule': rule_id, 'e': e}
            raise exceptions.CommandError(msg)


class ListNetworkQosRule(command.Lister):
    _description = _("List Network QoS rules")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_policy',
            metavar='<qos-policy>',
            help=_('QoS policy that contains the rule (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        columns = (
            'id',
            'qos_policy_id',
            'type',
            'max_kbps',
            'max_burst_kbps',
            'min_kbps',
            'min_kpps',
            'dscp_mark',
            'direction',
        )
        column_headers = (
            'ID',
            'QoS Policy ID',
            'Type',
            'Max Kbps',
            'Max Burst Kbits',
            'Min Kbps',
            'Min Kpps',
            'DSCP mark',
            'Direction',
        )
        qos = client.find_qos_policy(
            parsed_args.qos_policy, ignore_missing=False
        )

        return (
            column_headers,
            (utils.get_dict_properties(s, columns) for s in qos.rules),
        )


class SetNetworkQosRule(common.NeutronCommandWithExtraArgs):
    _description = _("Set Network QoS rule properties")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_policy',
            metavar='<qos-policy>',
            help=_('QoS policy that contains the rule (name or ID)'),
        )
        parser.add_argument(
            'id',
            metavar='<rule-id>',
            help=_('Network QoS rule to set (ID)'),
        )
        _add_rule_arguments(parser)
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        try:
            qos = network_client.find_qos_policy(
                parsed_args.qos_policy, ignore_missing=False
            )
            rule_type = _find_rule_type(qos, parsed_args.id)
            if not rule_type:
                raise Exception('Rule not found')
            attrs = _get_attrs(network_client, parsed_args)
            attrs.update(
                self._parse_extra_properties(parsed_args.extra_properties)
            )
            qos_id = attrs.pop('qos_policy_id')
            qos_rule = _rule_action_call(
                network_client, ACTION_FIND, rule_type
            )(attrs.pop('id'), qos_id)
            _rule_action_call(network_client, ACTION_SET, rule_type)(
                qos_rule, qos_id, **attrs
            )
        except Exception as e:
            msg = _('Failed to set Network QoS rule ID "%(rule)s": %(e)s') % {
                'rule': parsed_args.id,
                'e': e,
            }
            raise exceptions.CommandError(msg)


class ShowNetworkQosRule(command.ShowOne):
    _description = _("Display Network QoS rule details")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'qos_policy',
            metavar='<qos-policy>',
            help=_('QoS policy that contains the rule (name or ID)'),
        )
        parser.add_argument(
            'id',
            metavar='<rule-id>',
            help=_('Network QoS rule to show (ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        network_client = self.app.client_manager.network
        rule_id = parsed_args.id
        try:
            qos = network_client.find_qos_policy(
                parsed_args.qos_policy, ignore_missing=False
            )
            rule_type = _find_rule_type(qos, rule_id)
            if not rule_type:
                raise Exception('Rule not found')
            obj = _rule_action_call(network_client, ACTION_SHOW, rule_type)(
                rule_id, qos.id
            )
        except Exception as e:
            msg = _('Failed to show Network QoS rule ID "%(rule)s": %(e)s') % {
                'rule': rule_id,
                'e': e,
            }
            raise exceptions.CommandError(msg)
        display_columns, columns = _get_columns(obj)
        data = utils.get_item_properties(obj, columns)
        return display_columns, data



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import fixtures

from openstackclient.tests.functional import base


class HelpTests(base.TestCase):
    """Functional tests for openstackclient help output."""

    SERVER_COMMANDS = [
        ('server add security group', 'Add security group(s) to server'),
        ('server add volume', 'Add volume to server'),
        ('server backup create', 'Create a server backup image'),
        ('server create', 'Create a new server'),
        ('server delete', 'Delete server(s)'),
        ('server dump create', 'Create a dump file in server(s)'),
        (
            'server image create',
            'Create a new server disk image from an existing server',
        ),
        ('server list', 'List servers'),
        ('server lock', 'Lock server(s)'),
        ('server migrate', 'Migrate server to different host'),
        ('server pause', 'Pause server(s)'),
        ('server reboot', 'Perform a hard or soft server reboot'),
        ('server rebuild', 'Rebuild server'),
        ('server remove security group', 'Remove security group from server'),
        ('server remove volume', 'Remove volume from server'),
        ('server rescue', 'Put server in rescue mode'),
        ('server resize', 'Scale server to a new flavor'),
        ('server resume', 'Resume server(s)'),
        ('server set', 'Set server properties'),
        ('server shelve', 'Shelve and optionally offload server(s)'),
        ('server show', 'Show server details'),
        ('server ssh', 'SSH to server'),
        ('server start', 'Start server(s)'),
        ('server stop', 'Stop server(s)'),
        ('server suspend', 'Suspend server(s)'),
        ('server unlock', 'Unlock server(s)'),
        ('server unpause', 'Unpause server(s)'),
        ('server unrescue', 'Restore server from rescue mode'),
        ('server unset', 'Unset server properties'),
        ('server unshelve', 'Unshelve server(s)'),
    ]

    def test_server_commands_main_help(self):
        """Check server commands in main help message."""
        raw_output = self.openstack('help')
        for command, description in self.SERVER_COMMANDS:
            msg = f'Command: {command} not found in help output:\n{raw_output}'
            self.assertIn(command, raw_output, msg)
            msg = f'Description: {description} not found in help output:\n{raw_output}'
            self.assertIn(description, raw_output, msg)

    def test_server_only_help(self):
        """Check list of server-related commands only."""
        raw_output = self.openstack('help server')
        for command in [row[0] for row in self.SERVER_COMMANDS]:
            self.assertIn(command, raw_output)

    def test_networking_commands_help(self):
        """Check networking related commands in help message."""
        raw_output = self.openstack('help network list')
        self.assertIn('List networks', raw_output)
        raw_output = self.openstack('network create --help')
        self.assertIn('Create new network', raw_output)

    def test_commands_help_no_auth(self):
        """Check help commands without auth info."""
        # Pop all auth info. os.environ will be changed in loop, so do not
        # replace os.environ.keys() to os.environ
        for key in os.environ.keys():
            if key.startswith('OS_'):
                self.useFixture(fixtures.EnvironmentVariable(key, None))

        raw_output = self.openstack('help')
        self.assertIn('usage: openstack', raw_output)
        raw_output = self.openstack('--help')
        self.assertIn('usage: openstack', raw_output)

        raw_output = self.openstack('help network list')
        self.assertIn('List networks', raw_output)
        raw_output = self.openstack('network list --help')
        self.assertIn('List networks', raw_output)

        raw_output = self.openstack('help volume list')
        self.assertIn('List volumes', raw_output)
        raw_output = self.openstack('volume list --help')
        self.assertIn('List volumes', raw_output)

        raw_output = self.openstack('help server list')
        self.assertIn('List servers', raw_output)
        raw_output = self.openstack('server list --help')
        self.assertIn('List servers', raw_output)

        raw_output = self.openstack('help user list')
        self.assertIn('List users', raw_output)
        raw_output = self.openstack('user list --help')
        self.assertIn('List users', raw_output)

        raw_output = self.openstack('help image list')
        self.assertIn('List available images', raw_output)
        raw_output = self.openstack('image list --help')
        self.assertIn('List available images', raw_output)

        raw_output = self.openstack('help container list')
        self.assertIn('List containers', raw_output)
        raw_output = self.openstack('container list --help')
        self.assertIn('List containers', raw_output)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import uuid

from openstackclient.tests.functional.volume.v3 import common


class VolumeTypeTests(common.BaseVolumeTests):
    """Functional tests for volume type."""

    def test_volume_type_create_list(self):
        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume type create --private ' + name,
            parse_output=True,
        )
        self.addCleanup(
            self.openstack,
            'volume type delete ' + name,
        )
        self.assertEqual(name, cmd_output['name'])

        cmd_output = self.openstack(
            f'volume type show {name}',
            parse_output=True,
        )
        self.assertEqual(name, cmd_output['name'])

        cmd_output = self.openstack('volume type list', parse_output=True)
        self.assertIn(name, [t['Name'] for t in cmd_output])

        cmd_output = self.openstack(
            'volume type list --default',
            parse_output=True,
        )
        self.assertEqual(1, len(cmd_output))
        self.assertEqual('lvmdriver-1', cmd_output[0]['Name'])

    def test_volume_type_set_unset_properties(self):
        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume type create --private ' + name,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'volume type delete ' + name)
        self.assertEqual(name, cmd_output['name'])

        raw_output = self.openstack(
            f'volume type set --property a=b --property c=d {name}'
        )
        self.assertEqual("", raw_output)
        cmd_output = self.openstack(
            f'volume type show {name}',
            parse_output=True,
        )
        self.assertEqual({'a': 'b', 'c': 'd'}, cmd_output['properties'])

        raw_output = self.openstack(f'volume type unset --property a {name}')
        self.assertEqual("", raw_output)
        cmd_output = self.openstack(
            f'volume type show {name}',
            parse_output=True,
        )
        self.assertEqual({'c': 'd'}, cmd_output['properties'])

    def test_volume_type_set_unset_multiple_properties(self):
        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume type create --private ' + name,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'volume type delete ' + name)
        self.assertEqual(name, cmd_output['name'])

        raw_output = self.openstack(
            f'volume type set --property a=b --property c=d {name}'
        )
        self.assertEqual("", raw_output)
        cmd_output = self.openstack(
            f'volume type show {name}',
            parse_output=True,
        )
        self.assertEqual({'a': 'b', 'c': 'd'}, cmd_output['properties'])

        raw_output = self.openstack(
            f'volume type unset --property a --property c {name}'
        )
        self.assertEqual("", raw_output)
        cmd_output = self.openstack(
            f'volume type show {name}',
            parse_output=True,
        )
        self.assertEqual({}, cmd_output['properties'])

    def test_volume_type_set_unset_project(self):
        name = uuid.uuid4().hex
        cmd_output = self.openstack(
            'volume type create --private ' + name,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'volume type delete ' + name)
        self.assertEqual(name, cmd_output['name'])

        raw_output = self.openstack(f'volume type set --project admin {name}')
        self.assertEqual("", raw_output)

        raw_output = self.openstack(
            f'volume type unset --project admin {name}'
        )
        self.assertEqual("", raw_output)

    def test_multi_delete(self):
        vol_type1 = uuid.uuid4().hex
        vol_type2 = uuid.uuid4().hex
        self.openstack(f'volume type create {vol_type1}')
        time.sleep(5)
        self.openstack(f'volume type create {vol_type2}')
        time.sleep(5)
        cmd = f'volume type delete {vol_type1} {vol_type2}'
        raw_output = self.openstack(cmd)
        self.assertOutput('', raw_output)

    # NOTE: Add some basic functional tests with the old format to
    #       make sure the command works properly, need to change
    #       these to new test format when beef up all tests for
    #       volume type commands.
    def test_encryption_type(self):
        name = uuid.uuid4().hex
        encryption_type = uuid.uuid4().hex
        # test create new encryption type
        cmd_output = self.openstack(
            'volume type create '
            '--encryption-provider LuksEncryptor '
            '--encryption-cipher aes-xts-plain64 '
            '--encryption-key-size 128 '
            '--encryption-control-location front-end ' + encryption_type,
            parse_output=True,
        )
        expected = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'key_size': 128,
            'control_location': 'front-end',
        }
        for attr, value in expected.items():
            self.assertEqual(value, cmd_output['encryption'][attr])
        # test show encryption type
        cmd_output = self.openstack(
            'volume type show --encryption-type ' + encryption_type,
            parse_output=True,
        )
        expected = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'key_size': 128,
            'control_location': 'front-end',
        }
        for attr, value in expected.items():
            self.assertEqual(value, cmd_output['encryption'][attr])
        # test list encryption type
        cmd_output = self.openstack(
            'volume type list --encryption-type',
            parse_output=True,
        )
        encryption_output = [
            t['Encryption'] for t in cmd_output if t['Name'] == encryption_type
        ][0]
        expected = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'key_size': 128,
            'control_location': 'front-end',
        }
        for attr, value in expected.items():
            self.assertEqual(value, encryption_output[attr])
        # test set existing encryption type
        raw_output = self.openstack(
            'volume type set '
            '--encryption-key-size 256 '
            '--encryption-control-location back-end ' + encryption_type
        )
        self.assertEqual('', raw_output)
        cmd_output = self.openstack(
            'volume type show --encryption-type ' + encryption_type,
            parse_output=True,
        )
        expected = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'key_size': 256,
            'control_location': 'back-end',
        }
        for attr, value in expected.items():
            self.assertEqual(value, cmd_output['encryption'][attr])
        # test set new encryption type
        cmd_output = self.openstack(
            'volume type create --private ' + name,
            parse_output=True,
        )
        self.addCleanup(
            self.openstack,
            'volume type delete ' + name,
        )
        self.assertEqual(name, cmd_output['name'])

        raw_output = self.openstack(
            'volume type set '
            '--encryption-provider LuksEncryptor '
            '--encryption-cipher aes-xts-plain64 '
            '--encryption-key-size 128 '
            '--encryption-control-location front-end ' + name
        )
        self.assertEqual('', raw_output)

        cmd_output = self.openstack(
            'volume type show --encryption-type ' + name,
            parse_output=True,
        )
        expected = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'key_size': 128,
            'control_location': 'front-end',
        }
        for attr, value in expected.items():
            self.assertEqual(value, cmd_output['encryption'][attr])
        # test unset encryption type
        raw_output = self.openstack(
            'volume type unset --encryption-type ' + name
        )
        self.assertEqual('', raw_output)
        cmd_output = self.openstack(
            'volume type show --encryption-type ' + name,
            parse_output=True,
        )
        self.assertEqual({}, cmd_output['encryption'])
        # test delete encryption type
        raw_output = self.openstack('volume type delete ' + encryption_type)
        self.assertEqual('', raw_output)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from tempest.lib import exceptions as tempest_exc

from openstackclient.tests.functional.network.v2 import common


class PortTests(common.NetworkTagTests):
    """Functional tests for port"""

    base_command = 'port'

    NAME = uuid.uuid4().hex
    NETWORK_NAME = uuid.uuid4().hex

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if cls.haz_network:
            cls.NAME = uuid.uuid4().hex
            cls.NETWORK_NAME = uuid.uuid4().hex

            # Create a network for the port tests
            cls.openstack(f'network create {cls.NETWORK_NAME}')

    @classmethod
    def tearDownClass(cls):
        try:
            if cls.haz_network:
                raw_output = cls.openstack(
                    f'network delete {cls.NETWORK_NAME}'
                )
                cls.assertOutput('', raw_output)
        finally:
            super().tearDownClass()

    def test_port_delete(self):
        """Test create, delete multiple"""
        json_output = self.openstack(
            f'port create --network {self.NETWORK_NAME} {self.NAME}',
            parse_output=True,
        )
        id1 = json_output.get('id')
        self.assertIsNotNone(id1)
        self.assertIsNotNone(json_output.get('mac_address'))
        self.assertEqual(self.NAME, json_output.get('name'))

        json_output = self.openstack(
            f'port create --network {self.NETWORK_NAME} {self.NAME}x',
            parse_output=True,
        )
        id2 = json_output.get('id')
        self.assertIsNotNone(id2)
        self.assertIsNotNone(json_output.get('mac_address'))
        self.assertEqual(self.NAME + 'x', json_output.get('name'))

        # Clean up after ourselves
        raw_output = self.openstack(f'port delete {id1} {id2}')
        self.assertOutput('', raw_output)

    def test_port_list(self):
        """Test create defaults, list, delete"""
        json_output = self.openstack(
            f'port create --network {self.NETWORK_NAME} {self.NAME}',
            parse_output=True,
        )
        id1 = json_output.get('id')
        self.assertIsNotNone(id1)
        mac1 = json_output.get('mac_address')
        self.assertIsNotNone(mac1)
        self.addCleanup(self.openstack, f'port delete {id1}')
        self.assertEqual(self.NAME, json_output.get('name'))

        # sg for port2
        sg_name1 = uuid.uuid4().hex
        json_output = self.openstack(
            f'security group create {sg_name1}',
            parse_output=True,
        )
        sg_id1 = json_output.get('id')
        self.addCleanup(self.openstack, f'security group delete {sg_id1}')
        json_output = self.openstack(
            f'port create --network {self.NETWORK_NAME} '
            f'--security-group {sg_name1} {self.NAME}x',
            parse_output=True,
        )

        id2 = json_output.get('id')
        self.assertIsNotNone(id2)
        mac2 = json_output.get('mac_address')
        self.assertIsNotNone(mac2)
        self.addCleanup(self.openstack, f'port delete {id2}')
        self.assertEqual(self.NAME + 'x', json_output.get('name'))

        # Test list
        json_output = self.openstack(
            'port list',
            parse_output=True,
        )
        item_map = {
            item.get('ID'): item.get('MAC Address') for item in json_output
        }
        self.assertIn(id1, item_map.keys())
        self.assertIn(id2, item_map.keys())
        self.assertIn(mac1, item_map.values())
        self.assertIn(mac2, item_map.values())

        # Test list --long
        json_output = self.openstack(
            'port list --long',
            parse_output=True,
        )
        id_list = [item.get('ID') for item in json_output]
        self.assertIn(id1, id_list)
        self.assertIn(id2, id_list)
        item_sg_map = {
            item.get('ID'): item.get('Security Groups') for item in json_output
        }
        self.assertIn(id1, item_sg_map.keys())
        self.assertIn(id2, item_sg_map.keys())
        self.assertIn([sg_id1], item_sg_map.values())

        # Test list --mac-address
        json_output = self.openstack(
            f'port list --mac-address {mac2}',
            parse_output=True,
        )
        item_map = {
            item.get('ID'): item.get('MAC Address') for item in json_output
        }
        self.assertNotIn(id1, item_map.keys())
        self.assertIn(id2, item_map.keys())
        self.assertNotIn(mac1, item_map.values())
        self.assertIn(mac2, item_map.values())

        # Test list --security-group
        json_output = self.openstack(
            f'port list --security-group {sg_id1}',
            parse_output=True,
        )
        item_map = {
            item.get('ID'): item.get('Security Groups') for item in json_output
        }
        self.assertNotIn(id1, item_map.keys())
        self.assertIn(id2, item_map.keys())

        # Test list with unknown fields
        json_output = self.openstack(
            'port list -c ID -c Name -c device_id',
            parse_output=True,
        )
        id_list = [p['ID'] for p in json_output]
        self.assertIn(id1, id_list)
        self.assertIn(id2, id_list)
        # Check an unknown field does not exist
        self.assertNotIn('device_id', json_output[0])

        # Test list with only unknown fields
        exc = self.assertRaises(
            tempest_exc.CommandFailed,
            self.openstack,
            'port list -c device_id',
        )
        self.assertIn("No recognized column names in ['device_id']", str(exc))

    def test_port_set(self):
        """Test create, set, show, delete"""
        name = uuid.uuid4().hex
        json_output = self.openstack(
            'port create '
            f'--network {self.NETWORK_NAME} '
            '--description xyzpdq '
            f'--disable {name}',
            parse_output=True,
        )
        id1 = json_output.get('id')
        self.addCleanup(self.openstack, f'port delete {id1}')
        self.assertEqual(name, json_output.get('name'))
        self.assertEqual('xyzpdq', json_output.get('description'))
        self.assertEqual(False, json_output.get('admin_state_up'))

        raw_output = self.openstack(f'port set --enable {name}')
        self.assertOutput('', raw_output)

        json_output = self.openstack(
            f'port show {name}',
            parse_output=True,
        )
        sg_id = json_output.get('security_group_ids')[0]

        self.assertEqual(name, json_output.get('name'))
        self.assertEqual('xyzpdq', json_output.get('description'))
        self.assertEqual(True, json_output.get('admin_state_up'))
        self.assertIsNotNone(json_output.get('mac_address'))

        raw_output = self.openstack(
            f'port unset --security-group {sg_id} {id1}'
        )
        self.assertOutput('', raw_output)

        json_output = self.openstack(
            f'port show {name}',
            parse_output=True,
        )
        self.assertEqual([], json_output.get('security_group_ids'))

    def test_port_admin_set(self):
        """Test create, set (as admin), show, delete"""
        json_output = self.openstack(
            f'port create --network {self.NETWORK_NAME} {self.NAME}',
            parse_output=True,
        )
        id_ = json_output.get('id')
        self.addCleanup(self.openstack, f'port delete {id_}')

        raw_output = self.openstack(
            '--os-username admin '
            f'port set --mac-address 11:22:33:44:55:66 {self.NAME}'
        )
        self.assertOutput('', raw_output)
        json_output = self.openstack(
            f'port show {self.NAME}',
            parse_output=True,
        )
        self.assertEqual(json_output.get('mac_address'), '11:22:33:44:55:66')

    def test_port_set_sg(self):
        """Test create, set, show, delete"""
        sg_name1 = uuid.uuid4().hex
        json_output = self.openstack(
            f'security group create {sg_name1}',
            parse_output=True,
        )
        sg_id1 = json_output.get('id')
        self.addCleanup(self.openstack, f'security group delete {sg_id1}')

        sg_name2 = uuid.uuid4().hex
        json_output = self.openstack(
            f'security group create {sg_name2}',
            parse_output=True,
        )
        sg_id2 = json_output.get('id')
        self.addCleanup(self.openstack, f'security group delete {sg_id2}')

        name = uuid.uuid4().hex
        json_output = self.openstack(
            'port create '
            f'--network {self.NETWORK_NAME} '
            f'--security-group {sg_name1} {name}',
            parse_output=True,
        )
        id1 = json_output.get('id')
        self.addCleanup(self.openstack, f'port delete {id1}')
        self.assertEqual(name, json_output.get('name'))
        self.assertEqual([sg_id1], json_output.get('security_group_ids'))

        raw_output = self.openstack(
            f'port set --security-group {sg_name2} {name}'
        )
        self.assertOutput('', raw_output)

        json_output = self.openstack(
            f'port show {name}',
            parse_output=True,
        )
        self.assertEqual(name, json_output.get('name'))
        # NOTE(amotoki): The order of the field is not predictable,
        self.assertIsInstance(json_output.get('security_group_ids'), list)
        self.assertEqual(
            sorted([sg_id1, sg_id2]),
            sorted(json_output.get('security_group_ids')),
        )

        raw_output = self.openstack(
            f'port unset --security-group {sg_id1} {id1}'
        )
        self.assertOutput('', raw_output)

        json_output = self.openstack(
            f'port show {name}',
            parse_output=True,
        )
        self.assertEqual([sg_id2], json_output.get('security_group_ids'))

    def _create_resource_for_tag_test(self, name, args):
        return self.openstack(
            f'{self.base_command} create --network {self.NETWORK_NAME} {args} {name}',
            parse_output=True,
        )

    def _trunk_creation(self):
        pport = uuid.uuid4().hex
        sport1 = uuid.uuid4().hex
        sport2 = uuid.uuid4().hex
        trunk = uuid.uuid4().hex
        json_output = self.openstack(
            f'port create --network {self.NETWORK_NAME} {pport}',
            parse_output=True,
        )
        pport_id = json_output.get('id')
        json_output = self.openstack(
            f'port create --network {self.NETWORK_NAME} {sport1}',
            parse_output=True,
        )
        sport1_id = json_output.get('id')
        json_output = self.openstack(
            f'port create --network {self.NETWORK_NAME} {sport2}',
            parse_output=True,
        )
        sport2_id = json_output.get('id')

        self.openstack(
            f'network trunk create --parent-port {pport} {trunk}',
        )
        self.openstack(
            f'network trunk set --subport port={sport1},'
            f'segmentation-type=vlan,segmentation-id=100 {trunk}',
        )
        self.openstack(
            f'network trunk set --subport port={sport2},'
            f'segmentation-type=vlan,segmentation-id=101 {trunk}',
        )

        # NOTE(ralonsoh): keep this order to first delete the trunk and then
        # the ports.
        self.addCleanup(self.openstack, f'port delete {pport_id}')
        self.addCleanup(self.openstack, f'port delete {sport1_id}')
        self.addCleanup(self.openstack, f'port delete {sport2_id}')
        self.addCleanup(self.openstack, f'network trunk delete {trunk}')

        return pport_id, sport1_id, sport2_id

    def check_subports(self, subports, pport_id, sport1_id, sport2_id):
        self.assertEqual(2, len(subports))
        for subport in subports:
            if subport['port_id'] == sport1_id:
                self.assertEqual(100, subport['segmentation_id'])
            elif subport['port_id'] == sport2_id:
                self.assertEqual(101, subport['segmentation_id'])
            else:
                self.fail(
                    f'Port {pport_id} does not have subport '
                    f'{subport["port_id"]}'
                )
            self.assertEqual('vlan', subport['segmentation_type'])

    def test_port_list_with_trunk(self):
        pport_id, sport1_id, sport2_id = self._trunk_creation()

        # List all ports with "--long" flag to retrieve the trunk details
        json_output = self.openstack(
            'port list --long',
            parse_output=True,
        )
        port = next(port for port in json_output if port['ID'] == pport_id)
        subports = port['Trunk subports']
        self.check_subports(subports, pport_id, sport1_id, sport2_id)

    def test_port_show_with_trunk(self):
        pport_id, sport1_id, sport2_id = self._trunk_creation()

        # List all ports with "--long" flag to retrieve the trunk details
        port = self.openstack(
            f'port show {pport_id}',
            parse_output=True,
        )
        subports = port['trunk_details']['sub_ports']
        self.check_subports(subports, pport_id, sport1_id, sport2_id)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Compute v2 API Library

A collection of wrappers for deprecated Compute v2 APIs that are not
intentionally supported by SDK. Most of these are proxy APIs.
"""

import http

from openstack import exceptions as sdk_exceptions
from osc_lib import exceptions


# security groups


def create_security_group(compute_client, name=None, description=None):
    """Create a new security group

    https://docs.openstack.org/api-ref/compute/#create-security-group

    :param compute_client: A compute client
    :param str name: Security group name
    :param str description: Security group description
    :returns: A security group object
    """
    data = {
        'name': name,
        'description': description,
    }
    response = compute_client.post(
        '/os-security-groups', data=data, microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['security_group']


def list_security_groups(compute_client, all_projects=None):
    """Get all security groups

    https://docs.openstack.org/api-ref/compute/#list-security-groups

    :param compute_client: A compute client
    :param bool all_projects: If true, list from all projects
    :returns: A list of security group objects
    """
    url = '/os-security-groups'
    if all_projects is not None:
        url += f'?all_tenants={all_projects}'
    response = compute_client.get(url, microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    return response.json()['security_groups']


def find_security_group(compute_client, name_or_id):
    """Find the security group for a given name or ID

    https://docs.openstack.org/api-ref/compute/#show-security-group-details

    :param compute_client: A compute client
    :param name_or_id: The name or ID of the security group to look up
    :returns: A security group object
    :raises exception.NotFound: If a matching security group could not be
        found or more than one match was found
    """
    response = compute_client.get(
        f'/os-security-groups/{name_or_id}', microversion='2.1'
    )
    if response.status_code != http.HTTPStatus.NOT_FOUND:
        # there might be other, non-404 errors
        sdk_exceptions.raise_from_response(response)
        return response.json()['security_group']

    response = compute_client.get('/os-security-groups', microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    found = None
    security_groups = response.json()['security_groups']
    for security_group in security_groups:
        if security_group['name'] == name_or_id:
            if found:
                raise exceptions.NotFound(
                    f'multiple matches found for {name_or_id}'
                )
            found = security_group

    if not found:
        raise exceptions.NotFound(f'{name_or_id} not found')

    return found


def update_security_group(
    compute_client, security_group_id, name=None, description=None
):
    """Update an existing security group

    https://docs.openstack.org/api-ref/compute/#update-security-group

    :param compute_client: A compute client
    :param str security_group_id: The ID of the security group to update
    :param str name: Security group name
    :param str description: Security group description
    :returns: A security group object
    """
    data = {}
    if name:
        data['name'] = name
    if description:
        data['description'] = description
    response = compute_client.put(
        f'/os-security-groups/{security_group_id}',
        data=data,
        microversion='2.1',
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['security_group']


def delete_security_group(compute_client, security_group_id=None):
    """Delete a security group

    https://docs.openstack.org/api-ref/compute/#delete-security-group

    :param compute_client: A compute client
    :param str security_group_id: Security group ID
    :returns: None
    """
    response = compute_client.delete(
        f'/os-security-groups/{security_group_id}', microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)


# security group rules


def create_security_group_rule(
    compute_client,
    security_group_id=None,
    ip_protocol=None,
    from_port=None,
    to_port=None,
    remote_ip=None,
    remote_group=None,
):
    """Create a new security group rule

    https://docs.openstack.org/api-ref/compute/#create-security-group-rule

    :param compute_client: A compute client
    :param str security_group_id: Security group ID
    :param str ip_protocol: IP protocol, 'tcp', 'udp' or 'icmp'
    :param int from_port: Source port
    :param int to_port: Destination port
    :param str remote_ip: Source IP address in CIDR notation
    :param str remote_group: Remote security group
    :returns: A security group object
    """
    data = {
        'parent_group_id': security_group_id,
        'ip_protocol': ip_protocol,
        'from_port': from_port,
        'to_port': to_port,
        'cidr': remote_ip,
        'group_id': remote_group,
    }
    response = compute_client.post(
        '/os-security-group-rules', data=data, microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['security_group_rule']


def delete_security_group_rule(compute_client, security_group_rule_id=None):
    """Delete a security group rule

    https://docs.openstack.org/api-ref/compute/#delete-security-group-rule

    :param compute_client: A compute client
    :param str security_group_rule_id: Security group rule ID
    :returns: None
    """
    response = compute_client.delete(
        f'/os-security-group-rules/{security_group_rule_id}',
        microversion='2.1',
    )
    sdk_exceptions.raise_from_response(response)


# networks


def create_network(compute_client, name, subnet, share_subnet=None):
    """Create a new network

    https://docs.openstack.org/api-ref/compute/#create-network

    :param compute_client: A compute client
    :param str name: Network label
    :param int subnet: Subnet for IPv4 fixed addresses in CIDR notation
    :param bool share_subnet: Shared subnet between projects
    :returns: A network object
    """
    data = {
        'label': name,
        'cidr': subnet,
    }
    if share_subnet is not None:
        data['share_address'] = share_subnet

    response = compute_client.post(
        '/os-networks', data=data, microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['network']


def list_networks(compute_client):
    """Get all networks

    https://docs.openstack.org/api-ref/compute/#list-networks

    :param compute_client: A compute client
    :returns: A list of network objects
    """
    response = compute_client.get('/os-networks', microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    return response.json()['networks']


def find_network(compute_client, name_or_id):
    """Find the network for a given name or ID

    https://docs.openstack.org/api-ref/compute/#show-network-details

    :param compute_client: A compute client
    :param name_or_id: The name or ID of the network to look up
    :returns: A network object
    :raises exception.NotFound: If a matching network could not be found or
        more than one match was found
    """
    response = compute_client.get(
        f'/os-networks/{name_or_id}', microversion='2.1'
    )
    if response.status_code != http.HTTPStatus.NOT_FOUND:
        # there might be other, non-404 errors
        sdk_exceptions.raise_from_response(response)
        return response.json()['network']

    response = compute_client.get('/os-networks', microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    found = None
    networks = response.json()['networks']
    for network in networks:
        if network['label'] == name_or_id:
            if found:
                raise exceptions.NotFound(
                    f'multiple matches found for {name_or_id}'
                )
            found = network

    if not found:
        raise exceptions.NotFound(f'{name_or_id} not found')

    return found


def delete_network(compute_client, network_id):
    """Delete a network

    https://docs.openstack.org/api-ref/compute/#delete-network

    :param compute_client: A compute client
    :param string network_id: The network ID
    :returns: None
    """
    response = compute_client.delete(
        f'/os-networks/{network_id}', microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)


# floating ips


def create_floating_ip(compute_client, network):
    """Create a new floating ip

    https://docs.openstack.org/api-ref/compute/#create-allocate-floating-ip-address

    :param network: Name of floating IP pool
    :returns: A floating IP object
    """
    response = compute_client.post(
        '/os-floating-ips', data={'pool': network}, microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['floating_ip']


def list_floating_ips(compute_client):
    """Get all floating IPs

    https://docs.openstack.org/api-ref/compute/#list-floating-ip-addresses

    :returns: A list of floating IP objects
    """
    response = compute_client.get('/os-floating-ips', microversion='2.1')
    sdk_exceptions.raise_from_response(response)
    return response.json()['floating_ips']


def get_floating_ip(compute_client, floating_ip_id):
    """Get a floating IP

    https://docs.openstack.org/api-ref/compute/#show-floating-ip-address-details

    :param string floating_ip_id: The floating IP address
    :returns: A floating IP object
    """
    response = compute_client.get(
        f'/os-floating-ips/{floating_ip_id}', microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)
    return response.json()['floating_ip']


def delete_floating_ip(compute_client, floating_ip_id):
    """Delete a floating IP

    https://docs.openstack.org/api-ref/compute/#delete-deallocate-floating-ip-address

    :param string floating_ip_id: The floating IP address
    :returns: None
    """
    response = compute_client.delete(
        f'/os-floating-ips/{floating_ip_id}', microversion='2.1'
    )
    sdk_exceptions.raise_from_response(response)


# floating ip pools


def list_floating_ip_pools(compute_client):
    """Get all floating IP pools

    https://docs.openstack.org/api-ref/compute/#list-floating-ip-pools

    :param compute_client: A compute client
    :returns: A list of floating IP pool objects
    """
    response = compute_client.get('/os-floating-ip-pools', microversion='2.1')
    sdk_exceptions.raise_from_response(response)

    return response.json()['floating_ip_pools']



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock
from unittest.mock import call

from osc_lib.cli import format_columns
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils as tests_utils
from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import volume_type


class TestType(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.volume_types_mock = self.volume_client.volume_types
        self.volume_types_mock.reset_mock()

        self.volume_type_access_mock = self.volume_client.volume_type_access
        self.volume_type_access_mock.reset_mock()

        self.volume_encryption_types_mock = (
            self.volume_client.volume_encryption_types
        )
        self.volume_encryption_types_mock.reset_mock()

        self.projects_mock = self.identity_client.projects
        self.projects_mock.reset_mock()


class TestTypeCreate(TestType):
    def setUp(self):
        super().setUp()

        self.new_volume_type = volume_fakes.create_one_volume_type(
            methods={'set_keys': None},
        )
        self.project = identity_fakes.FakeProject.create_one_project()
        self.columns = (
            'description',
            'id',
            'is_public',
            'name',
        )
        self.data = (
            self.new_volume_type.description,
            self.new_volume_type.id,
            True,
            self.new_volume_type.name,
        )

        self.volume_types_mock.create.return_value = self.new_volume_type
        self.projects_mock.get.return_value = self.project
        # Get the command object to test
        self.cmd = volume_type.CreateVolumeType(self.app, None)

    def test_type_create_public(self):
        arglist = [
            "--description",
            self.new_volume_type.description,
            "--public",
            self.new_volume_type.name,
        ]
        verifylist = [
            ("description", self.new_volume_type.description),
            ("is_public", True),
            ("name", self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.create.assert_called_with(
            self.new_volume_type.name,
            description=self.new_volume_type.description,
            is_public=True,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_type_create_private(self):
        arglist = [
            "--description",
            self.new_volume_type.description,
            "--private",
            "--project",
            self.project.id,
            self.new_volume_type.name,
        ]
        verifylist = [
            ("description", self.new_volume_type.description),
            ("is_public", False),
            ("project", self.project.id),
            ("name", self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.create.assert_called_with(
            self.new_volume_type.name,
            description=self.new_volume_type.description,
            is_public=False,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_type_create_with_properties(self):
        arglist = [
            '--property',
            'myprop=myvalue',
            # this combination isn't viable server-side but is okay for testing
            '--multiattach',
            '--cacheable',
            '--replicated',
            '--availability-zone',
            'az1',
            self.new_volume_type.name,
        ]
        verifylist = [
            ('properties', {'myprop': 'myvalue'}),
            ('multiattach', True),
            ('cacheable', True),
            ('replicated', True),
            ('availability_zones', ['az1']),
            ('name', self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.create.assert_called_with(
            self.new_volume_type.name, description=None
        )
        self.new_volume_type.set_keys.assert_called_once_with(
            {
                'myprop': 'myvalue',
                'multiattach': '<is> True',
                'cacheable': '<is> True',
                'replication_enabled': '<is> True',
                'RESKEY:availability_zones': 'az1',
            }
        )

        self.columns += ('properties',)
        self.data += (format_columns.DictColumn(None),)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_public_type_create_with_project_public(self):
        arglist = [
            '--project',
            self.project.id,
            self.new_volume_type.name,
        ]
        verifylist = [
            ('is_public', None),
            ('project', self.project.id),
            ('name', self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )

    def test_type_create_with_encryption(self):
        encryption_info = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'key_size': '128',
            'control_location': 'front-end',
        }
        encryption_type = volume_fakes.create_one_encryption_volume_type(
            attrs=encryption_info,
        )
        self.new_volume_type = volume_fakes.create_one_volume_type(
            attrs={'encryption': encryption_info},
        )
        self.volume_types_mock.create.return_value = self.new_volume_type
        self.volume_encryption_types_mock.create.return_value = encryption_type
        encryption_columns = (
            'description',
            'encryption',
            'id',
            'is_public',
            'name',
        )
        encryption_data = (
            self.new_volume_type.description,
            format_columns.DictColumn(encryption_info),
            self.new_volume_type.id,
            True,
            self.new_volume_type.name,
        )
        arglist = [
            '--encryption-provider',
            'LuksEncryptor',
            '--encryption-cipher',
            'aes-xts-plain64',
            '--encryption-key-size',
            '128',
            '--encryption-control-location',
            'front-end',
            self.new_volume_type.name,
        ]
        verifylist = [
            ('encryption_provider', 'LuksEncryptor'),
            ('encryption_cipher', 'aes-xts-plain64'),
            ('encryption_key_size', 128),
            ('encryption_control_location', 'front-end'),
            ('name', self.new_volume_type.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.create.assert_called_with(
            self.new_volume_type.name,
            description=None,
        )
        body = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'key_size': 128,
            'control_location': 'front-end',
        }
        self.volume_encryption_types_mock.create.assert_called_with(
            self.new_volume_type,
            body,
        )
        self.assertEqual(encryption_columns, columns)
        self.assertCountEqual(encryption_data, data)


class TestTypeDelete(TestType):
    volume_types = volume_fakes.create_volume_types(count=2)

    def setUp(self):
        super().setUp()

        self.volume_types_mock.get = volume_fakes.get_volume_types(
            self.volume_types,
        )
        self.volume_types_mock.delete.return_value = None

        # Get the command object to mock
        self.cmd = volume_type.DeleteVolumeType(self.app, None)

    def test_type_delete(self):
        arglist = [self.volume_types[0].id]
        verifylist = [("volume_types", [self.volume_types[0].id])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_types_mock.delete.assert_called_with(self.volume_types[0])
        self.assertIsNone(result)

    def test_delete_multiple_types(self):
        arglist = []
        for t in self.volume_types:
            arglist.append(t.id)
        verifylist = [
            ('volume_types', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for t in self.volume_types:
            calls.append(call(t))
        self.volume_types_mock.delete.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_types_with_exception(self):
        arglist = [
            self.volume_types[0].id,
            'unexist_type',
        ]
        verifylist = [
            ('volume_types', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self.volume_types[0], exceptions.CommandError]
        with mock.patch.object(
            utils, 'find_resource', side_effect=find_mock_result
        ) as find_mock:
            try:
                self.cmd.take_action(parsed_args)
                self.fail('CommandError should be raised.')
            except exceptions.CommandError as e:
                self.assertEqual(
                    '1 of 2 volume types failed to delete.', str(e)
                )
            find_mock.assert_any_call(
                self.volume_types_mock, self.volume_types[0].id
            )
            find_mock.assert_any_call(self.volume_types_mock, 'unexist_type')

            self.assertEqual(2, find_mock.call_count)
            self.volume_types_mock.delete.assert_called_once_with(
                self.volume_types[0]
            )


class TestTypeList(TestType):
    volume_types = volume_fakes.create_volume_types()

    columns = [
        "ID",
        "Name",
        "Is Public",
    ]
    columns_long = columns + ["Description"]
    data_with_default_type = [(volume_types[0].id, volume_types[0].name, True)]
    data = []
    for t in volume_types:
        data.append(
            (
                t.id,
                t.name,
                t.is_public,
            )
        )
    data_long = []
    for t in volume_types:
        data_long.append(
            (
                t.id,
                t.name,
                t.is_public,
                t.description,
            )
        )

    def setUp(self):
        super().setUp()

        self.volume_types_mock.list.return_value = self.volume_types
        self.volume_types_mock.default.return_value = self.volume_types[0]
        # get the command to test
        self.cmd = volume_type.ListVolumeType(self.app, None)

    def test_type_list_without_options(self):
        arglist = []
        verifylist = [
            ("long", False),
            ("is_public", None),
            ("default", False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.list.assert_called_once_with(is_public=None)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_type_list_with_options(self):
        arglist = [
            "--long",
            "--public",
        ]
        verifylist = [
            ("long", True),
            ("is_public", True),
            ("default", False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.list.assert_called_once_with(is_public=True)
        self.assertEqual(self.columns_long, columns)
        self.assertCountEqual(self.data_long, list(data))

    def test_type_list_with_private_option(self):
        arglist = [
            "--private",
        ]
        verifylist = [
            ("long", False),
            ("is_public", False),
            ("default", False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.list.assert_called_once_with(is_public=False)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_type_list_with_default_option(self):
        arglist = [
            "--default",
        ]
        verifylist = [
            ("encryption_type", False),
            ("long", False),
            ("is_public", None),
            ("default", True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.default.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data_with_default_type, list(data))

    def test_type_list_with_encryption(self):
        encryption_type = volume_fakes.create_one_encryption_volume_type(
            attrs={'volume_type_id': self.volume_types[0].id},
        )
        encryption_info = {
            'provider': 'LuksEncryptor',
            'cipher': None,
            'key_size': None,
            'control_location': 'front-end',
        }
        encryption_columns = self.columns + [
            "Encryption",
        ]
        encryption_data = []
        encryption_data.append(
            (
                self.volume_types[0].id,
                self.volume_types[0].name,
                self.volume_types[0].is_public,
                volume_type.EncryptionInfoColumn(
                    self.volume_types[0].id,
                    {self.volume_types[0].id: encryption_info},
                ),
            )
        )
        encryption_data.append(
            (
                self.volume_types[1].id,
                self.volume_types[1].name,
                self.volume_types[1].is_public,
                volume_type.EncryptionInfoColumn(self.volume_types[1].id, {}),
            )
        )

        self.volume_encryption_types_mock.list.return_value = [encryption_type]
        arglist = [
            "--encryption-type",
        ]
        verifylist = [
            ("encryption_type", True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_encryption_types_mock.list.assert_called_once_with()
        self.volume_types_mock.list.assert_called_once_with(is_public=None)
        self.assertEqual(encryption_columns, columns)
        self.assertCountEqual(encryption_data, list(data))


class TestTypeSet(TestType):
    def setUp(self):
        super().setUp()

        self.project = identity_fakes.FakeProject.create_one_project()
        self.projects_mock.get.return_value = self.project

        self.volume_type = volume_fakes.create_one_volume_type(
            methods={'set_keys': None},
        )
        self.volume_types_mock.get.return_value = self.volume_type
        self.volume_encryption_types_mock.create.return_value = None
        self.volume_encryption_types_mock.update.return_value = None

        self.cmd = volume_type.SetVolumeType(self.app, None)

    def test_type_set(self):
        arglist = [
            '--name',
            'new_name',
            '--description',
            'new_description',
            '--private',
            self.volume_type.id,
        ]
        verifylist = [
            ('name', 'new_name'),
            ('description', 'new_description'),
            ('properties', None),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        kwargs = {
            'name': 'new_name',
            'description': 'new_description',
            'is_public': False,
        }
        self.volume_types_mock.update.assert_called_with(
            self.volume_type.id, **kwargs
        )
        self.assertIsNone(result)

        self.volume_type_access_mock.add_project_access.assert_not_called()
        self.volume_encryption_types_mock.update.assert_not_called()
        self.volume_encryption_types_mock.create.assert_not_called()

    def test_type_set_property(self):
        arglist = [
            '--property',
            'myprop=myvalue',
            # this combination isn't viable server-side but is okay for testing
            '--multiattach',
            '--cacheable',
            '--replicated',
            '--availability-zone',
            'az1',
            self.volume_type.id,
        ]
        verifylist = [
            ('name', None),
            ('description', None),
            ('properties', {'myprop': 'myvalue'}),
            ('multiattach', True),
            ('cacheable', True),
            ('replicated', True),
            ('availability_zones', ['az1']),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_type.set_keys.assert_called_once_with(
            {
                'myprop': 'myvalue',
                'multiattach': '<is> True',
                'cacheable': '<is> True',
                'replication_enabled': '<is> True',
                'RESKEY:availability_zones': 'az1',
            }
        )
        self.volume_type_access_mock.add_project_access.assert_not_called()
        self.volume_encryption_types_mock.update.assert_not_called()
        self.volume_encryption_types_mock.create.assert_not_called()

    def test_type_set_with_empty_project(self):
        arglist = [
            '--project',
            '',
            self.volume_type.id,
        ]
        verifylist = [
            ('project', ''),
            ('volume_type', self.volume_type.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_type.set_keys.assert_not_called()
        self.volume_type_access_mock.add_project_access.assert_not_called()
        self.volume_encryption_types_mock.update.assert_not_called()
        self.volume_encryption_types_mock.create.assert_not_called()

    def test_type_set_with_project(self):
        arglist = [
            '--project',
            self.project.id,
            self.volume_type.id,
        ]
        verifylist = [
            ('project', self.project.id),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_type.set_keys.assert_not_called()
        self.volume_type_access_mock.add_project_access.assert_called_with(
            self.volume_type.id,
            self.project.id,
        )
        self.volume_encryption_types_mock.update.assert_not_called()
        self.volume_encryption_types_mock.create.assert_not_called()

    def test_type_set_with_new_encryption(self):
        self.volume_encryption_types_mock.update.side_effect = (
            exceptions.NotFound('NotFound')
        )
        arglist = [
            '--encryption-provider',
            'LuksEncryptor',
            '--encryption-cipher',
            'aes-xts-plain64',
            '--encryption-key-size',
            '128',
            '--encryption-control-location',
            'front-end',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_provider', 'LuksEncryptor'),
            ('encryption_cipher', 'aes-xts-plain64'),
            ('encryption_key_size', 128),
            ('encryption_control_location', 'front-end'),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        body = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'key_size': 128,
            'control_location': 'front-end',
        }
        self.volume_encryption_types_mock.update.assert_called_with(
            self.volume_type,
            body,
        )
        self.volume_encryption_types_mock.create.assert_called_with(
            self.volume_type,
            body,
        )

    @mock.patch.object(utils, 'find_resource')
    def test_type_set_with_existing_encryption(self, mock_find):
        mock_find.side_effect = [self.volume_type, "existing_encryption_type"]
        arglist = [
            '--encryption-provider',
            'LuksEncryptor',
            '--encryption-cipher',
            'aes-xts-plain64',
            '--encryption-control-location',
            'front-end',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_provider', 'LuksEncryptor'),
            ('encryption_cipher', 'aes-xts-plain64'),
            ('encryption_control_location', 'front-end'),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_type.set_keys.assert_not_called()
        self.volume_type_access_mock.add_project_access.assert_not_called()
        body = {
            'provider': 'LuksEncryptor',
            'cipher': 'aes-xts-plain64',
            'control_location': 'front-end',
        }
        self.volume_encryption_types_mock.update.assert_called_with(
            self.volume_type,
            body,
        )
        self.volume_encryption_types_mock.create.assert_not_called()

    def test_type_set_new_encryption_without_provider(self):
        self.volume_encryption_types_mock.update.side_effect = (
            exceptions.NotFound('NotFound')
        )
        arglist = [
            '--encryption-cipher',
            'aes-xts-plain64',
            '--encryption-key-size',
            '128',
            '--encryption-control-location',
            'front-end',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_cipher', 'aes-xts-plain64'),
            ('encryption_key_size', 128),
            ('encryption_control_location', 'front-end'),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )
        self.assertEqual(
            "Command Failed: One or more of the operations failed",
            str(exc),
        )

        self.volume_type.set_keys.assert_not_called()
        self.volume_type_access_mock.add_project_access.assert_not_called()
        body = {
            'cipher': 'aes-xts-plain64',
            'key_size': 128,
            'control_location': 'front-end',
        }
        self.volume_encryption_types_mock.update.assert_called_with(
            self.volume_type,
            body,
        )
        self.volume_encryption_types_mock.create.assert_not_called()


class TestTypeShow(TestType):
    columns = (
        'access_project_ids',
        'description',
        'id',
        'is_public',
        'name',
        'properties',
    )

    def setUp(self):
        super().setUp()

        self.volume_type = volume_fakes.create_one_volume_type()
        self.data = (
            None,
            self.volume_type.description,
            self.volume_type.id,
            True,
            self.volume_type.name,
            format_columns.DictColumn(self.volume_type.extra_specs),
        )

        self.volume_types_mock.get.return_value = self.volume_type

        # Get the command object to test
        self.cmd = volume_type.ShowVolumeType(self.app, None)

    def test_type_show(self):
        arglist = [self.volume_type.id]
        verifylist = [
            ("encryption_type", False),
            ("volume_type", self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.get.assert_called_with(self.volume_type.id)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_type_show_with_access(self):
        arglist = [self.volume_type.id]
        verifylist = [("volume_type", self.volume_type.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        private_type = volume_fakes.create_one_volume_type(
            attrs={'is_public': False},
        )
        type_access_list = volume_fakes.create_one_type_access()
        with mock.patch.object(
            self.volume_types_mock,
            'get',
            return_value=private_type,
        ):
            with mock.patch.object(
                self.volume_type_access_mock,
                'list',
                return_value=[type_access_list],
            ):
                columns, data = self.cmd.take_action(parsed_args)
                self.volume_types_mock.get.assert_called_once_with(
                    self.volume_type.id
                )
                self.volume_type_access_mock.list.assert_called_once_with(
                    private_type.id
                )

        self.assertEqual(self.columns, columns)
        private_type_data = (
            format_columns.ListColumn([type_access_list.project_id]),
            private_type.description,
            private_type.id,
            private_type.is_public,
            private_type.name,
            format_columns.DictColumn(private_type.extra_specs),
        )
        self.assertCountEqual(private_type_data, data)

    def test_type_show_with_list_access_exec(self):
        arglist = [self.volume_type.id]
        verifylist = [("volume_type", self.volume_type.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        private_type = volume_fakes.create_one_volume_type(
            attrs={'is_public': False},
        )
        with mock.patch.object(
            self.volume_types_mock, 'get', return_value=private_type
        ):
            with mock.patch.object(
                self.volume_type_access_mock, 'list', side_effect=Exception()
            ):
                columns, data = self.cmd.take_action(parsed_args)
                self.volume_types_mock.get.assert_called_once_with(
                    self.volume_type.id
                )
                self.volume_type_access_mock.list.assert_called_once_with(
                    private_type.id
                )

        self.assertEqual(self.columns, columns)
        private_type_data = (
            None,
            private_type.description,
            private_type.id,
            private_type.is_public,
            private_type.name,
            format_columns.DictColumn(private_type.extra_specs),
        )
        self.assertCountEqual(private_type_data, data)

    def test_type_show_with_encryption(self):
        encryption_type = volume_fakes.create_one_encryption_volume_type()
        encryption_info = {
            'provider': 'LuksEncryptor',
            'cipher': None,
            'key_size': None,
            'control_location': 'front-end',
        }
        self.volume_type = volume_fakes.create_one_volume_type(
            attrs={'encryption': encryption_info},
        )
        self.volume_types_mock.get.return_value = self.volume_type
        self.volume_encryption_types_mock.get.return_value = encryption_type
        encryption_columns = (
            'access_project_ids',
            'description',
            'encryption',
            'id',
            'is_public',
            'name',
            'properties',
        )
        encryption_data = (
            None,
            self.volume_type.description,
            format_columns.DictColumn(encryption_info),
            self.volume_type.id,
            True,
            self.volume_type.name,
            format_columns.DictColumn(self.volume_type.extra_specs),
        )
        arglist = ['--encryption-type', self.volume_type.id]
        verifylist = [
            ('encryption_type', True),
            ("volume_type", self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_types_mock.get.assert_called_with(self.volume_type.id)
        self.volume_encryption_types_mock.get.assert_called_with(
            self.volume_type.id
        )
        self.assertEqual(encryption_columns, columns)
        self.assertCountEqual(encryption_data, data)


class TestTypeUnset(TestType):
    project = identity_fakes.FakeProject.create_one_project()
    volume_type = volume_fakes.create_one_volume_type(
        methods={'unset_keys': None},
    )

    def setUp(self):
        super().setUp()

        self.volume_types_mock.get.return_value = self.volume_type

        # Return a project
        self.projects_mock.get.return_value = self.project

        # Get the command object to test
        self.cmd = volume_type.UnsetVolumeType(self.app, None)

    def test_type_unset(self):
        arglist = [
            '--property',
            'property',
            '--property',
            'multi_property',
            self.volume_type.id,
        ]
        verifylist = [
            ('properties', ['property', 'multi_property']),
            ('volume_type', self.volume_type.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.volume_type.unset_keys.assert_called_once_with(
            ['property', 'multi_property']
        )
        self.assertIsNone(result)

    def test_type_unset_project_access(self):
        arglist = [
            '--project',
            self.project.id,
            self.volume_type.id,
        ]
        verifylist = [
            ('project', self.project.id),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.volume_type_access_mock.remove_project_access.assert_called_with(
            self.volume_type.id,
            self.project.id,
        )

    def test_type_unset_not_called_without_project_argument(self):
        arglist = [
            '--project',
            '',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_type', False),
            ('project', ''),
            ('volume_type', self.volume_type.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)
        self.volume_encryption_types_mock.delete.assert_not_called()
        self.assertFalse(
            self.volume_type_access_mock.remove_project_access.called
        )

    def test_type_unset_failed_with_missing_volume_type_argument(self):
        arglist = [
            '--project',
            'identity_fakes.project_id',
        ]
        verifylist = [
            ('project', 'identity_fakes.project_id'),
        ]

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_type_unset_encryption_type(self):
        arglist = [
            '--encryption-type',
            self.volume_type.id,
        ]
        verifylist = [
            ('encryption_type', True),
            ('volume_type', self.volume_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.volume_encryption_types_mock.delete.assert_called_with(
            self.volume_type
        )
        self.assertIsNone(result)


class TestColumns(TestType):
    def test_encryption_info_column_with_info(self):
        fake_volume_type = volume_fakes.create_one_volume_type()
        type_id = fake_volume_type.id

        encryption_info = {
            'provider': 'LuksEncryptor',
            'cipher': None,
            'key_size': None,
            'control_location': 'front-end',
        }
        col = volume_type.EncryptionInfoColumn(
            type_id, {type_id: encryption_info}
        )
        self.assertEqual(
            utils.format_dict(encryption_info), col.human_readable()
        )
        self.assertEqual(encryption_info, col.machine_readable())

    def test_encryption_info_column_without_info(self):
        fake_volume_type = volume_fakes.create_one_volume_type()
        type_id = fake_volume_type.id

        col = volume_type.EncryptionInfoColumn(type_id, {})
        self.assertEqual('-', col.human_readable())
        self.assertIsNone(col.machine_readable())
